from wsurf.geolab.optimization.sparse_mkl._mkl_interface import (MKL, sparse_matrix_t, _create_mkl_sparse,
                                           _export_mkl, _order_mkl_handle, _destroy_mkl_handle, _type_check,
                                           _get_numpy_layout, _convert_to_csr, _empty_output_check, LAYOUT_CODE_C,
                                           _out_matrix, _check_return_value, debug_print)

import scipy.sparse as _sps
import ctypes as _ctypes
import numpy as np


def _gram_matrix_sparse(matrix_a, aat=False, reorder_output=False):
    """
    Calculate the gram matrix aTa for sparse matrix and return a sparse matrix

    :param matrix_a: Sparse matrix
    :type matrix_a: scipy.sparse.csr_matrix, scipy.sparse.csc_matrix
    :param aat: Return A (dot) AT instead of AT (dot) A
    :type aat: bool
    :param reorder_output:
    :type reorder_output: bool
    :return: Sparse matrix
    :rtype: scipy.sparse.csr_matrix
    """

    sp_ref_a, double_prec = _create_mkl_sparse(matrix_a)

    if _sps.isspmatrix_csc(matrix_a):
        sp_ref_a = _convert_to_csr(sp_ref_a)

    _order_mkl_handle(sp_ref_a)

    ref_handle = sparse_matrix_t()

    ret_val = MKL._mkl_sparse_syrk(10 if aat else 11,
                                   sp_ref_a,
                                   _ctypes.byref(ref_handle))

    # Check return
    _check_return_value(ret_val, "mkl_sparse_syrk")

    if reorder_output:
        _order_mkl_handle(ref_handle)

    output_arr = _export_mkl(ref_handle, double_prec, output_type="csr")

    _destroy_mkl_handle(sp_ref_a)
    _destroy_mkl_handle(ref_handle)

    return output_arr


def _gram_matrix_sparse_to_dense(matrix_a, aat=False, scalar=1., out=None, out_scalar=None):
    """
    Calculate the gram matrix aTa for sparse matrix and return a dense matrix

    :param matrix_a: Sparse matrix
    :type matrix_a: scipy.sparse.csr_matrix, scipy.sparse.csc_matrix
    :param aat: Return A (dot) AT instead of AT (dot) A
    :type aat: bool
    :param scalar: Multiply output by a scalar value
    :type scalar: float
    :return: Dense matrix
    :rtype: numpy.ndarray
    """

    sp_ref_a, double_prec = _create_mkl_sparse(matrix_a)

    if _sps.isspmatrix_csc(matrix_a):
        sp_ref_a = _convert_to_csr(sp_ref_a, destroy_original=True)

    _order_mkl_handle(sp_ref_a)

    out_dtype = np.float64 if double_prec else np.float32
    output_ctype = _ctypes.c_double if double_prec else _ctypes.c_float
    out_dim = matrix_a.shape[0] if aat else matrix_a.shape[1]

    output_arr = _out_matrix((out_dim, out_dim), out_dtype, order="C", out_arr=out)
    _, output_ld = _get_numpy_layout(output_arr)

    if _empty_output_check(matrix_a, matrix_a):
        return output_arr

    func = MKL._mkl_sparse_d_syrkd if double_prec else MKL._mkl_sparse_s_syrkd

    ret_val = func(10 if aat else 11,
                   sp_ref_a,
                   scalar,
                   float(out_scalar) if out_scalar is not None else 1.,
                   output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)),
                   LAYOUT_CODE_C,
                   output_ld)

    # Check return
    _check_return_value(ret_val, func.__name__)

    _destroy_mkl_handle(sp_ref_a)

    # This fixes a specific bug in mkl_sparse_d_syrkd which returns a full matrix
    # This stupid thing only happens with specific flags
    # I could probably leave it but it's pretty annoying
    if not aat and out is None:
        output_arr[np.tril_indices(output_arr.shape[0], k=-1)] = 0.

    return output_arr


def _gram_matrix_dense_to_dense(matrix_a, aat=False, scalar=1., out=None, out_scalar=None):
    """
    Calculate the gram matrix aTa for dense matrix and return a dense matrix

    :param matrix_a: Dense matrix
    :type matrix_a: numpy.ndarray
    :param aat: Return A (dot) AT instead of AT (dot) A
    :type aat: bool
    :param scalar: Multiply output by a scalar value
    :type scalar: float
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: Dense matrix
    :rtype: numpy.ndarray
    """

    # Get dimensions
    n, k = matrix_a.shape if aat else matrix_a.shape[::-1]

    # Get the memory order for arrays
    layout_a, ld_a = _get_numpy_layout(matrix_a)
    double_precision = matrix_a.dtype == np.float64

    # Set the MKL function for precision
    func = MKL._cblas_dsyrk if double_precision else MKL._cblas_ssyrk
    output_ctype = _ctypes.c_double if double_precision else _ctypes.c_float

    # Allocate an array for outputs and set functions and types for float or doubles
    output_arr = _out_matrix((n, n), matrix_a.dtype, order="C" if layout_a == LAYOUT_CODE_C else "F", out_arr=out)

    func(layout_a,
         121,
         111 if aat else 112,
         n,
         k,
         scalar,
         matrix_a,
         ld_a,
         float(out_scalar) if out_scalar is not None else 1.,
         output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)),
         n)

    return output_arr


def _gram_matrix(matrix, transpose=False, cast=False, dense=False, reorder_output=False, out=None, out_scalar=None):
    """
    Calculate a gram matrix (AT (dot) A) from a sparse matrix.

    :param matrix: Sparse matrix (CSR format is required but will convert if cast is True)
    :type matrix: scipy.sparse.csr_matrix
    :param transpose: Calculate A (dot) AT instead
    :type transpose: bool
    :param cast: Make internal copies to convert matrix to a float matrix or convert to a CSR matrix if necessary
    :type cast: bool
    :param dense: Produce a dense matrix output instead of a sparse matrix
    :type dense: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: Gram matrix
    :rtype: scipy.sparse.csr_matrix, np.ndarray
    """

    # Check for edge condition inputs which result in empty outputs
    if _empty_output_check(matrix, matrix):
        debug_print("Skipping multiplication because AT (dot) A must yield an empty matrix")
        output_shape = (matrix.shape[1], matrix.shape[1]) if transpose else (matrix.shape[0], matrix.shape[0])
        output_func = _sps.csr_matrix if _sps.isspmatrix(matrix) else np.zeros
        return output_func(output_shape, dtype=matrix.dtype)

    matrix = _type_check(matrix, cast=cast)

    if _sps.isspmatrix(matrix) and not (_sps.isspmatrix_csr(matrix) or _sps.isspmatrix_csc(matrix)):
        raise ValueError("gram_matrix requires sparse matrix to be CSR or CSC format")
    if _sps.isspmatrix_csc(matrix) and not cast:
        raise ValueError("gram_matrix cannot use a CSC matrix unless cast=True")
    elif not _sps.isspmatrix(matrix):
        return _gram_matrix_dense_to_dense(matrix, aat=transpose, out=out, out_scalar=out_scalar)
    elif dense:
        return _gram_matrix_sparse_to_dense(matrix, aat=transpose,  out=out, out_scalar=out_scalar)
    elif out is not None:
        raise ValueError("out argument cannot be used with sparse (dot) sparse matrix multiplication")
    else:
        return _gram_matrix_sparse(matrix, aat=transpose, reorder_output=reorder_output)





#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from traits.api import Button, Str, Float, Bool, Int

from traits.api import on_trait_change

from traitsui.api import View, Item, HGroup, Group, VGroup, Tabbed

import numpy as np

# ------------------------------------------------------------------------------

from .geolab.gui.geolabcomponent import GeolabComponent

from .bspline_optimizer import WsurfOptimizer

from .geolab.geometry import Polyline

from .geolab.plotter.plotutilities import uv_wheel_lut_table

# ------------------------------------------------------------------------------

__author__ = 'Davide Pellis'


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
#                            Optimization GUI
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


class WeingartenGUI(GeolabComponent):
    # --------------------------------------------------------------------------
    #                                 Traits
    # --------------------------------------------------------------------------

    name = Str('Weingarten B-spline')

    iterations = Int(1, label='Iterations')

    epsilon = Float(0.001, label='Dumping')

    step = Float(1, label='Step')

    step_control = Bool(True)

    alignment = Float(0.1)

    reference_closeness = Float(0.1)

    shape_preservation = Float(0)

    gliding_boundary = Float(0.01)

    control_fairness = Float(0.2)

    isolines_parallel = Float(0)

    fitting = Float(0, label='Linear')

    scale = Float(1)

    curve_fitting = Button(label='Fit')

    reparametrize_points = Button(label='Reparametrize')

    reinitialize = Button(label='Initialize')

    optimize = Button(label='Optimize')

    geometric_error = Str('_', label='Geometric')

    interactive = Bool(False, label='Interactive')

    boundary = Bool(False, label='boundary')

    u_sampling = Int(0, label='Sampling (U,V)')

    v_sampling = Int(0)

    u_degree = Int(0, label='Degree (U,V)')

    v_degree = Int(0)

    u_points = Int(0, label='Points (U,V)')

    v_points = Int(0)

    rebuild_button = Button(label='Rebuild')

    resample_button = Button(label='Resample')

    plot_samples_check = Bool(False, label='Show samples')

    plot_clusters_check = Bool(False, label='Plot clusters')

    fix_button = Button(label='Fix', tooltip='Fix selected vertices')

    unfix_button = Button(label='Unfix', tooltip='Unfix selected vertices')

    reference_button = Button(label='Reference', tooltip=
    'Set the current shape as reference')

    color_view_check = Bool(False, label='Color')

    hk_view_check = Bool(True, label='H-K')

    k1k2_view_check = Bool(False, label='k1-k2')

    optimizer_values_check = Bool(False, label='Unknowns', tooltip=
    'Show values from unknowns vector')

    K_coeff = Float(0, label='a')

    H_coeff = Float(0, label='b')

    C_coeff = Float(0, label='c')

    linear_relation = Bool(True)

    # --------------------------------------------------------------------------
    #                              Component View
    # --------------------------------------------------------------------------

    view = View(
        VGroup(Group('alignment',
                     'fitting',
                     'reference_closeness',
                     'shape_preservation',
                     'isolines_parallel',
                     'control_fairness',
                     'gliding_boundary',
                     'boundary',
                     show_border=True,
                     label='Weights'),
               HGroup('K_coeff',
                      'H_coeff',
                      'C_coeff',
                      show_border=True,
                      label='Linear relation: aK + bH + c = 0'),
               VGroup(HGroup(Item('u_sampling', width=-100),
                             Item('v_sampling', show_label=False,
                                  width=-100), ),
                      HGroup(Item('u_points', width=-100),
                             Item('v_points', show_label=False,
                                  width=-100), ),
                      HGroup(Item('u_degree', width=-100),
                             Item('v_degree', show_label=False,
                                  width=-100), ),
                      HGroup('plot_samples_check',
                             '_',
                             Item('rebuild_button', show_label=False),
                             Item('resample_button', show_label=False),
                             ),
                      show_border=True,
                      label='B-spline'),
               VGroup(
                   HGroup('step', 'step_control',
                          show_border=True),
                   HGroup(Item('interactive',
                               tooltip='Interactive optimization',
                               show_label=False, ),
                          Item('_'),
                          'reinitialize',
                          'optimize',
                          show_labels=False,
                          show_border=True),
                   show_border=True,
                   label='Optimizer'),
               VGroup('geometric_error',
                      style='readonly',
                      label='errors [ mean | max ]',
                      show_border=True, ),
               show_border=True,
               show_labels=True),
        resizable=True,
    )

    # -------------------------------------------------------------------------
    #                                Attributes
    # -------------------------------------------------------------------------
    '''Define here all component standard attributes'''

    surface_lut = None

    optimizer = WsurfOptimizer()

    __reinitialize = False

    # -------------------------------------------------------------------------
    #                                 Properties
    # -------------------------------------------------------------------------

    @property
    def bspline(self):
        return self.geolab.current_object.geometry

    @property
    def current_object(self):
        return self.geolab.current_object

    # -------------------------------------------------------------------------
    #                            Special Methods
    # -------------------------------------------------------------------------
    '''These are overwriting of special methods, fired by geolab'''

    def geolab_settings(self):
        self.geolab.add_scene('3D')
        position = [0.010216647233753599, -1.1378681444403156,
                    4.796698057325836, 0.0, 1.0, 0.0, 0.010216647233753599,
                    -1.1378681444403156, 0.0, 50.0, 4.302506863495564,
                    5.430352877034375, False]
        self.geolab.add_scene('2D', interaction='locked',
                              position=position,
                              style='base')
        # self.geolab.add_scene('2D', position=position)
        self.geolab.height = 800
        self.geolab.width = 800
        self.geolab.side_width = 400

    def geolab_object_add(self):
        self.geolab.delete_selected_object()
        self.__reinitialize = True

    def geolab_object_added(self):
        obj = self.geolab.last_object
        bspline = obj.geometry
        self.optimizer.reset_scaling()
        P = bspline.number_of_control_points
        bspline.u_sampling = int(P[0] * 2)
        bspline.v_sampling = int(P[1] * 2)
        if self.__reinitialize:
            self.geolab.object_changed()
            self.__reinitialize = False
        
    def geolab_geometry_reset(self):
        P = self.bspline.number_of_control_points
        self.bspline.u_sampling = int(P[0] * 3)
        self.bspline.v_sampling = int(P[1] * 3)

    def geolab_object_change(self):
        self.current_object.surface_plot = 'color'

    def geolab_object_changed(self):
        if self.geolab.current_object is None:
            return
        self.optimizer.bspline = self.bspline
        self.current_object.update_size()
        self.u_degree = self.bspline.u_degree
        self.v_degree = self.bspline.v_degree
        P = self.bspline.number_of_control_points
        self.u_points = P[0]
        self.v_points = P[1]
        self.u_sampling = self.bspline.u_sampling
        self.v_sampling = self.bspline.v_sampling
        self.bspline.sampling = (self.u_sampling, self.v_sampling)
        P = self.bspline.points()
        self.surface_lut = uv_wheel_lut_table(P)
        self.optimizer.verbose = True
        self.geolab.current_object.update_size()
        self.optimizer.initialization()
        self.geolab.clear()
        self.initialize_bspline_plot_functions()
        self.plot_diagrams()
        self.plot_samples()
        self.plot_fixed_vertices()
        self.geolab.update_plot()

    def geolab_object_save(self, file_name):
        self.optimizer.resize_bspline()

    def geolab_object_saved(self, file_name):
        self.optimizer.unresize_bspline()

    def geolab_set_state(self, state):
        """Function called to set interactive states that should be
        switched off when other selections are enabled and vice-versa.
        Pre-set states - to do not use - are:
            - 'select_object'
            - 'split_edges':
            - 'collapse_edges':
            - 'move_vertices':
            - 'select_vertices':
            - 'select_faces':
            - 'select_edges':
            - 'select_boundary_vertices':
        """
        if state != 'bs-interactive':
            self.interactive = False

    def geolab_fix_vertices(self, selected):
        self.optimizer.fix_points(selected)

    def geolab_unfix_vertices(self, selected):
        self.optimizer.unfix_points(selected)

    def geolab_set_reference(self):
        self.optimizer.set_reference()

    # -------------------------------------------------------------------------
    #                              Plot Functions
    # -------------------------------------------------------------------------

    def initialize_bspline_plot_functions(self):
        obj = self.geolab.current_object

        def plot_surface_color():
            obj.plot_surface(color=self.surface_lut,
                             vertex_data=np.arange(len(self.surface_lut)))

        obj.add_surface_callback(plot_surface_color, name='surface rainbow')

        def plot_surface_hk_isolines():
            init = not self.optimizer_values_check
            if self.geolab.is_moving_vertex:
                K, H, k1, k2 = self.optimizer.bspline.curvatures()
            else:
                H = self.optimizer.mean(init)
                K = self.optimizer.gaussian(init)
            obj.plot_surface(vertex_data=H, color='r', iso_surface=True,
                             name='iso_h', tube_radius=0.2, shading=False)
            obj.plot_surface(vertex_data=K, color='b', iso_surface=True,
                             name='iso_k', tube_radius=0.2, shading=False)
            obj.plot_surface(color='w')

        obj.add_surface_callback(plot_surface_hk_isolines, name='HK isolines')

        def plot_surface_k1k2_isolines():
            init = not self.optimizer_values_check
            if self.geolab.is_moving_vertex:
                K, H, k1, k2 = self.optimizer.bspline.curvatures()
            else:
                k1, k2 = self.optimizer.principal_curvatures(init)
            obj.plot_surface(vertex_data=k1, color='cornflower',
                             iso_surface=True, name='iso_h', tube_radius=0.2,
                             shading=False)
            obj.plot_surface(vertex_data=k2, color='orange', iso_surface=True,
                             name='iso_k', tube_radius=0.2, shading=False)
            obj.plot_surface(color='w')

        obj.add_surface_callback(plot_surface_k1k2_isolines,
                                 name='k1-k2 isolines')

        obj.surface_plot = 'HK isolines'
        self.hk_view_check = True

    def plot_diagrams(self):

        def plot_diagrams_callback():
            init = not self.optimizer_values_check
            U = self.optimizer.bspline.u_sampling
            V = self.optimizer.bspline.v_sampling
            if self.geolab.is_moving_vertex:
                K, H, k1, k2 = self.optimizer.bspline.curvatures()
            else:
                k1, k2 = self.optimizer.principal_curvatures(init)
                H = self.optimizer.mean(init)
                K = self.optimizer.gaussian(init)
            Kmin = min(np.min(K), -0.05)
            Kmax = max(np.max(K), 0.05)
            Hmin = min(np.min(H), -0.05)
            Hmax = max(np.max(H), 0.05)
            k1min = min(np.min(k1), -0.05)
            k1max = max(np.max(k1), 0.05)
            k2min = min(np.min(k2), -0.05)
            k2max = max(np.max(k2), 0.05)
            cK = (Kmax + Kmin) / 2
            cH = (Hmax + Hmin) / 2
            ck1 = (k1max + k1min) / 2
            ck2 = (k2max + k2min) / 2
            Kc = K - cK
            Hc = H - cH
            k1c = k1 - ck1
            k2c = k2 - ck2
            Kn = np.max(abs(Kc))
            Hn = np.max(abs(Hc))
            k12n = max(np.max(np.abs(k1c)), np.max(np.abs(k2c)))
            HK = np.column_stack((Hc / Hn, Kc / Kn, np.zeros(U * V)))
            k1k2 = np.column_stack((k1c / k12n, k2c / k12n, np.zeros(U * V)))
            k1k2[:, 1] -= 2.2
            r = 0.012
            # C = [[-cH/Hn, -cK/Kn, 0],[-ck1/k12n, -ck2/k12n-2.2, 0]]
            Hax = (np.linspace(-10, 10, 201) - cH) / Hn
            Hax = Hax[Hax < 1.3]
            Hax = Hax[Hax > -1.3]
            Kax = (np.linspace(-10, 10, 201) - cK) / Kn
            Kax = Kax[Kax < 1.5]
            Kax = Kax[Kax > -1.2]
            Kax2 = (np.linspace(-10, 10, 2001) - cK) / Kn
            Kax2 = Kax2[Kax2 < 1.5]
            Kax2 = Kax2[Kax2 > -1.2]
            HKaxes = Axes(x=Hax, y=Kax, cx=-cH / Hn, cy=-cK / Kn)
            HKaxes2 = Axes(y=Kax2, cx=-cH / Hn, cy=-cK / Kn, d=0.01)
            k1ax = (np.linspace(-10, 10, 201) - ck1) / k12n
            k1ax = k1ax[k1ax < 1.3]
            k1ax = k1ax[k1ax > -1.3]
            k2ax = (np.linspace(-10, 10, 201) - ck2) / k12n - 2.2
            k2ax = k2ax[k2ax < -1.25]
            k2ax = k2ax[k2ax > -3.8]
            k1k2axes = Axes(x=k1ax, y=k2ax, cx=-ck1 / k12n, cy=-ck2 / k12n - 2.2)
            line = Polyline([[-3, -1.2, 0], [3, -1.2, 0]])
            self.geolab.plot_mesh_edges(HKaxes, color=(60, 60, 60), scene='2D',
                                        name='hk_axes')
            self.geolab.plot_mesh_edges(HKaxes2, color=(60, 60, 60), scene='2D',
                                        name='hk_axes2')
            self.geolab.plot_mesh_edges(k1k2axes, color=(60, 60, 60), scene='2D',
                                        name='k1k2_axes')

            self.geolab.plot_polyline(line, color=(240, 240, 240),
                                      line_width=6,
                                      scene='2D')
            if not self.plot_clusters_check:
                color = self.surface_lut
                self.geolab.plot_points(HK, radius=r,
                                        color=color,
                                        resolution=5,
                                        vertex_data=np.arange(len(HK)),
                                        shading=False,
                                        lut_range='-:+',
                                        scene='2D',
                                        name='HK')
                self.geolab.plot_points(k1k2, radius=r, color=color,
                                        resolution=5,
                                        vertex_data=np.arange(len(HK)),
                                        shading=False,
                                        lut_range='-:+',
                                        scene='2D',
                                        name='k1k2')
            else:
                data = self.optimizer.clusters
                self.geolab.plot_points(HK, radius=r,
                                        color='blue-red',
                                        resolution=5,
                                        vertex_data=data,
                                        shading=False,
                                        lut_range='-:+',
                                        scene='2D',
                                        name='HK')
                self.geolab.plot_points(k1k2,
                                        radius=r,
                                        color='blue-red',
                                        resolution=5,
                                        vertex_data=data,
                                        shading=False,
                                        lut_range='-:+',
                                        scene='2D',
                                        name='k1k2')

        self.geolab.current_object.add_plot_callback(plot_diagrams_callback)

    def plot_fixed_vertices(self):

        def plot_fixed_callback():
            v = self.optimizer.fixed_points
            self.geolab.current_object.plot_glyph(vertex_indices=v,
                                                  glyph_type='axes',
                                                  color='orange',
                                                  shading=False,
                                                  line_width=4,
                                                  scale_factor=0.3,
                                                  name='fixed')

        self.geolab.current_object.add_plot_callback(plot_fixed_callback)

    def plot_samples(self):

        def plot_samples_callback():
            if not self.plot_samples_check:
                self.geolab.current_object.remove('samples')
            else:
                P = self.bspline.points()
                self.geolab.current_object.plot_points(P,
                                                       color=self.surface_lut,
                                                       vertex_data=np.arange(len(P)),
                                                       lut_range='-:+',
                                                       radius=0.07,
                                                       shading=False,
                                                       name='samples')

        self.geolab.current_object.add_plot_callback(plot_samples_callback)

    @on_trait_change('plot_samples_check')
    def update_plot_fired(self):
        self.geolab.current_object.update_plot()

    # -------------------------------------------------------------------------
    #                              Settings
    # -------------------------------------------------------------------------

    def set_settings(self):
        self.optimizer.step_control = self.step_control
        self.optimizer.step = self.step
        self.optimizer.iterations = 1
        self.optimizer.epsilon = self.epsilon
        self.optimizer.step = self.step
        self.optimizer.include_boundary = self.boundary
        self.optimizer.linear_coefficients = (self.K_coeff, self.H_coeff,
                                              -self.C_coeff)
        self.optimizer.set_weight('linear_fitting', self.fitting)
        self.optimizer.set_weight('gliding', self.gliding_boundary)
        self.optimizer.set_weight('reference', self.reference_closeness)
        self.optimizer.set_weight('control_lengths', self.shape_preservation)
        self.optimizer.set_weight('hk', self.alignment)
        self.optimizer.set_weight('control_fairness', self.control_fairness)
        self.optimizer.set_weight('isolines_parallel', self.isolines_parallel)
        self.optimizer.adaptive_curve = False

    # -------------------------------------------------------------------------
    #                              Optimization
    # -------------------------------------------------------------------------

    def optimization_step(self):
        if not self.interactive:
            self.geolab.set_state(None)
        self.set_settings()
        self.optimizer.optimize()
        self.print_error()
        if True:
            if self.optimizer.geometric_error()[1] > 1:
                self.optimizer.reinitialize()
        self.current_object.update_plot()

    def print_error(self):
        self.geometric_error = self.optimizer.geometric_error_string()

    @on_trait_change('optimize')
    def optimize_fired(self):
        self.current_object.iterate(self.optimization_step, self.iterations)

    @on_trait_change('interactive')
    def interactive_optimize_fired(self):
        self.geolab.set_state('bs-interactive')
        if self.interactive:
            def start():
                self.optimizer.handle = self.current_object.selected_vertices

            def interact():
                self.current_object.iterate(self.optimization_step, 1)

            def end():
                self.current_object.iterate(self.optimization_step, 5)

            self.current_object.move_vertices(interact, start, end)
        else:
            self.optimizer.handle = None
            self.current_object.move_vertices_off()

    @on_trait_change('reinitialize')
    def reinitialize_fired(self):
        self.set_settings()
        self.optimizer.reinitialization()
        self.print_error()
        self.current_object.update_plot()

    # -------------------------------------------------------------------------
    #                              Traits
    # -------------------------------------------------------------------------

    @on_trait_change('rebuild_button')
    def rebuild_fired(self):
        if self.u_sampling < self.u_points:
            self.u_sampling = self.u_points + 1
        if self.v_sampling < self.v_points:
            self.v_sampling = self.v_points + 1
        self.bspline.u_sampling = self.u_sampling
        self.bspline.v_sampling = self.v_sampling
        self.bspline.rebuild(self.u_degree, self.v_degree, self.u_points,
                             self.v_points)
        self.geolab.object_changed()
        self.geolab.current_object.update_plot()

    @on_trait_change('resample_button')
    def resample_fired(self):
        self.bspline.u_sampling = self.u_sampling
        self.bspline.v_sampling = self.v_sampling
        self.geolab.object_changed()
        self.geolab.current_object.update_plot()

    def set_plot_state(self, state=None):
        if state != 'color':
            self.color_view_check = False
        if state != 'hk':
            self.hk_view_check = False
        if state != 'k1k2':
            self.k1k2_view_check = False

    @on_trait_change('color_view_check')
    def color_view_fired(self):
        if self.color_view_check:
            self.set_plot_state('color')
            self.geolab.current_object.surface_plot = 'surface rainbow'

    @on_trait_change('hk_view_check')
    def hk_view_fired(self):
        if self.hk_view_check:
            self.set_plot_state('hk')
            self.geolab.current_object.surface_plot = 'HK isolines'

    @on_trait_change('k1k2_view_check')
    def k1k2_view_fired(self):
        if self.k1k2_view_check:
            self.set_plot_state('k1k2')
            self.geolab.current_object.surface_plot = 'k1-k2 isolines'

    @on_trait_change('optimizer_values_check')
    def optimizer_values_fired(self):
        self.geolab.current_object.update_plot()

    @on_trait_change('scale')
    def scale_fired(self):
        self.optimizer.scale = self.scale
        self.current_object.update_size()
        self.optimizer.reinitialization()
        self.geolab.set_state(None)
        self.current_object.update_plot()


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class Axes(object):

    def __init__(self, x=[], y=[], cx=0, cy=0, d=0.03):
        self.d = d
        E = len(x) + len(y)
        self._Nx = len(x)
        self._Ny = len(y)
        self.V = E * 2
        self.E = E
        nx = len(x)
        ny = len(y)
        Xm = np.column_stack((x, np.repeat(cy - self.d, nx), np.zeros(nx)))
        Xp = np.column_stack((x, np.repeat(cy + self.d, nx), np.zeros(nx)))
        Ym = np.column_stack((np.repeat(cx - self.d, ny), y, np.zeros(ny)))
        Yp = np.column_stack((np.repeat(cx + self.d, ny), y, np.zeros(ny)))
        self.vertices = np.vstack((Xm, Ym, Xp, Yp))

    def edge_vertices(self):
        x = np.arange(self._Nx)
        y = np.arange(self._Ny) + self._Nx
        v1 = np.hstack((x, y))
        v2 = v1 + self.E
        return v1, v2


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


def WsurfGUI(file_name=None):

    from wsurf.geolab.gui.geolabgui import GeolabGUI

    component = WeingartenGUI()

    GUI = GeolabGUI()

    GUI.add_component(component)

    if file_name is not None:
        GUI.open_obj_file(file_name)

    GUI.start()



#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

# -----------------------------------------------------------------------------

'''_'''

__author__ = 'Davide Pellis'


def sum_repeated(array, field):
    field = field[:]
    imap = np.zeros(np.amax(field)+1, dtype=np.int)
    index = np.arange(array.shape[0])
    k, j = np.unique(field, True)
    imap[k] = np.arange(k.shape[0])
    result = array[j]
    field = np.delete(field,j)
    index = np.delete(index,j)
    while field.shape[0] > 0:
        _, j = np.unique(field, True)
        result[imap[field[j]]] += array[index[j]]
        field = np.delete(field,j)
        index = np.delete(index,j)
    return result


def repeated_range(array, offset=0):
    if array.shape[0] == 0:
        return np.array([])
    k = np.unique(array)
    imap = np.zeros(np.amax(k) + 1, dtype=np.int)
    imap[k] = np.arange(offset, offset + k.shape[0])
    rrange = imap[array]
    return rrange


def orthogonal_vectors(array):
    array = format_array(array)
    if len(array.shape) == 1:
        return orthogonal_vector(array)
    O = np.zeros(array.shape)
    O[:,0] = - array[:,1]
    O[:,1] = array[:,0]
    O[np.where((O[:,0] == 0) & (O[:,1] == 0))[0],1] = 1
    O = O / np.linalg.norm(O, axis=1, keepdims=True)
    return O


def orthogonal_vector(array):
    if array[0] == 0 and array[1] == 0:
            if array[2] == 0:
                raise ValueError('zero vector')
            return np.array([1,0,0])
    O = np.array([-array[1],array[0],0])
    return O / np.linalg.norm(O)


def normalize(array, axis=1):
    array = format_array(array)
    eps = 1e-10
    array = array / (np.linalg.norm(array, axis=axis, keepdims=True) + eps)
    return array


def remap(source, target_range=(0,1)):
    t_int = target_range[1] - target_range[0]
    s_min = np.min(source)
    s_max = np.max(source)
    s_int = s_max - s_min
    s = t_int/s_int
    remapped = (source - s_min) * s + target_range[0]
    return remapped


def format_array(array, shape=None):
    if array is None:
        return None
    if type(shape) is int:
        shape = (shape,)
    if type(array) == float or type(array) == int:
        array = np.array(array)
    elif type(array) == tuple:
        array = list(array)
    if type(array) == list:
        array = np.array(array)
    if shape is None or array.shape == shape:
        return array
    else:
        array_shape = list(array.shape)
        for i in range(len(shape)-len(array.shape)):
            array_shape = [1] + array_shape
        shape = np.array(shape)
        array_shape = np.array(array_shape)
        shape[shape == None] = array_shape[shape == None]
        repeats = shape - array_shape + 1
        array = np.tile(array, repeats)
        cmd = 'array['
        for n in shape:
            cmd += ':{},'.format(n)
        cmd += ']'
        array = eval(cmd)
    return array

            
def bounding_shape(arrays):
    shape = []
    for array in arrays:
        if type(array) != np.ndarray:
            array = np.array(array)
        for i in range(len(array.shape) - len(shape)):
            shape = [1] + shape
        for i in range(len(shape)):
            if array.shape[i] > shape[i]:
                shape[i] = array.shape[i]
    return shape

    
    
    
    
    
    
    
    
    
    
    
    
    
    


#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from __future__ import unicode_literals

import numpy as np

from mayavi.core.api import ModuleManager

# -----------------------------------------------------------------------------

from wsurf.geolab import utilities

from wsurf.geolab.plotter import colorutilities

'''_'''

__author__ = 'Davide Pellis'


# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------

def color_library():
    library = {

        'r': (255, 0, 0),

        'g': (0, 255, 0),

        'b': (0, 0, 255),

        'c': (0, 255, 255),

        'm': (255, 0, 255),

        'y': (255, 255, 0),

        'k': (0, 0, 0),

        'w': (255, 255, 255),

        'white': (255, 255, 255),

        'black': (0, 0, 0),

        'gray_40': (153, 153, 153),

        'gray_50': (128, 128, 128),

        'gray_60': (102, 102, 102),

        'magenta': (255, 0, 255),

        'cyan': (0, 255, 255),

        'azure': (0, 136, 255),

        'red_orange': (255, 120, 64),

        'orange': (255, 128, 0),

        'green': (0, 255, 128),

        'yellow': (255, 255, 0),

        'golden_yellow': (255, 221, 0),

        'metallic_gold': (212, 175, 55),

        'gold': (218, 165, 32),

        'aqua': (0, 255, 255),

        'cadet_blue': (95, 158, 160),

        'sandy_brown': (244, 164, 96),

        'white_blue': (204, 229, 255),

        'cornflower': (122, 163, 230),

        'chrome_yellow': (255, 170, 0),

        'safety_orange': (255, 119, 0),

        'carmine_red': (255, 0, 51),

        'violet': (153, 17, 170),

        'vivid_lime_green': (170, 221, 34),

        'smashed_pumpkin': (251, 109, 55),

        'icterine': (251, 247, 94),

        'tiffany_blue': (1, 198, 178),

        'sea_blue': (0, 108, 150),

        'prussian_blue': (0, 57, 89),

        'capri': (1, 190, 254),

        'amber': (255, 125, 0),

        'vivid_raspberry': (255, 0, 109),

        'spring_bud': (173, 255, 2),

        'electric_violet': (143, 0, 255),

        'cyber': (255, 211, 0),

    }
    return library


def palette_library():
    library = {

        'raimbow_1': ['w', 'azure', 'safety_orange', 'violet',
                      'vivid_lime_green', 'carmine_red', (255, 186, 0),
                      (0, 60, 95)],

        'one_more_raimbow': ['w', 'azure', 'safety_orange', 'violet',
                             'vivid_lime_green', 'carmine_red', 'chrome_yellow'],

        'lose_me_away': ['w', 'smashed_pumpkin', 'icterine', 'tiffany_blue',
                         'sea_blue', 'prussian_blue'],

        'happy_colors': ['w', 'capri', 'golden_yellow', 'amber',
                         'vivid_raspberry', 'spring_bud', 'electric_violet'],

        'milti_copat': ['w', (30, 56, 136), (245, 230, 99), (156, 56, 72),
                        (71, 168, 189), (255, 173, 105)],

        'overtly_cautions': ['w', (26, 91, 109), (216, 201, 155), (216, 151, 60),
                             (189, 99, 47), (39, 62, 71), (140, 179, 201),
                             (152, 150, 118)]

    }
    return library


def palette(key):
    try:
        palettes = palette_library()
        return palettes[key]
    except:
        return key


def map_color(key):
    library = color_library()
    try:
        color = library[key]
    except KeyError:
        if type(key) == tuple and len(key) == 3:
            color = key
        elif type(key) == str:
            color = (128, 128, 128)
        elif key == None:
            color = None
        else:
            color = (128, 128, 128)
    return color


def rgb_float_color(key):
    c = map_color(key)
    return (c[0] / 255, c[1] / 255, c[2] / 255)


def lut_table(colors, opacity):
    '''
    elif isinstance(colors, unicode):
        colors = [colors]
    '''
    if isinstance(colors, np.ndarray):
        return colors
    colors = palette(colors)
    if isinstance(colors, str):
        colors = [colors]
    elif isinstance(colors, tuple):
        colors = [colors]
    if isinstance(opacity, float):
        opacity = [opacity]
    if isinstance(opacity, int):
        opacity = [opacity]
    if len(colors) != len(opacity):
        opacity = np.repeat(opacity[0], len(colors))
    rgb = []
    for key in colors:
        rgb.append(map_color(key))
    lut = np.ones((256, 4)) * 255
    color = rgb[-1]
    alpha = opacity[-1]
    lut[:, 0] = color[0] * np.ones([256])
    lut[:, 1] = color[1] * np.ones([256])
    lut[:, 2] = color[2] * np.ones([256])
    lut[:, 3] = alpha * 255 * np.ones([256])
    L = len(rgb)
    N = 256 // L
    n = 0
    for i in range(L - 1):
        color = rgb[i]
        alpha = opacity[i]
        lut[n:n + N, 0] = color[0] * np.ones([N])
        lut[n:n + N, 1] = color[1] * np.ones([N])
        lut[n:n + N, 2] = color[2] * np.ones([N])
        lut[n:n + N, 3] = alpha * 255 * np.ones([N])
        n += N
    return lut


def bwr_lut_table(expansion=2):
    X = np.linspace(0, 1, 128)
    A = 255 * (X ** expansion)
    B = A[::-1]
    C = 255 * np.ones(128)
    lut = np.ones((256, 4)) * 255
    R = np.hstack((C, B))
    G = np.hstack((A, B))
    B = np.hstack((A, C))
    lut[:, 0] = R
    lut[:, 1] = G
    lut[:, 2] = B
    return lut


def rgb_lut_table(expansion=2, opacity=1):
    X = np.linspace(0, 1, 64)
    increase = 255 * (X ** (expansion ** (-1)))
    decrease = increase[::-1]
    one = 255 * np.ones(64)
    zero = np.zeros(64)
    lut = np.ones((256, 4)) * 255
    B = np.hstack((one, decrease, zero, zero))
    G = np.hstack((increase, one, one, decrease))
    R = np.hstack((zero, zero, increase, one))
    lut[:, 0] = R
    lut[:, 1] = G
    lut[:, 2] = B
    lut[:, 3] = int(255 * opacity)
    return lut


def rgbm_lut_table(expansion=2):
    X = np.linspace(0, 1, 64)
    increase = 255 * (X ** (expansion ** (-1)))
    decrease = increase[::-1]
    one = 255 * np.ones(64)
    zero = np.zeros(64)
    lut = np.ones((256, 4)) * 255
    B = np.hstack((one, decrease, zero, zero))
    G = np.hstack((increase, one, one, decrease))
    R = np.hstack((zero, zero, increase, one))
    lut[:, 0] = R
    lut[:, 1] = G
    lut[:, 2] = B
    lut[-1, :] = [255, 0, 255, 255]
    return lut


def ramp(values, slopes=1):
    counter = np.sign(slopes)
    slopes = abs(slopes)
    val = 0
    if counter < 0:
        val = 1
    vmax = np.max(values)
    vmin = np.min(values)
    start = vmin
    domain = vmax - vmin
    if domain == 0:
        domain = 1e-5
    step = domain / slopes
    out = np.zeros(len(values))
    for i in range(slopes):
        mask = np.logical_and(values <= start + step, values >= start)
        out[mask] = counter * (values[mask] - start) / step + val
        counter *= -1
        start += step
        if val == 0:
            val = 1
        else:
            val = 0
    # out /= np.max(out)
    return out


def xyz_lut_table(points, variation=0):
    if variation == 0:
        a = b = c = 1
    else:
        a = variation + 1
        b = variation
        c = -variation
    C = (np.max(points, axis=0) + np.min(points, axis=0)) / 2
    x = 255 * ramp((points[:, 0] - C[0]) * 1.01, a)
    y = 255 * ramp((points[:, 1] - C[1]) * 1.01, -b)
    z = 255 * ramp((points[:, 2] - C[2]) * 1.01, c)
    a = 255 * np.ones(len(points))
    lut = np.column_stack((x, y, z, a))
    lut = np.array(np.round(lut), 'i')
    return lut


def uv_wheel_lut_table(points):
    P = points[:, [0, 1]]
    C = (np.max(P, axis=0) + np.min(P, axis=0)) / 2
    P[:, 0] -= C[0]
    P[:, 1] -= C[1]
    P[:, 0] /= np.max(np.abs(P[:, 0]))
    P[:, 1] /= np.max(np.abs(P[:, 1]))
    r = np.max(np.abs(P), axis=1)
    v = utilities.normalize(P, axis=1)
    a = np.degrees(np.arccos(v[:, 0]))
    a[v[:, 1] < 0] = 360 - a[v[:, 1] < 0]
    rgb = hsv_to_rgb(np.column_stack((a, r, np.ones(len(r)))))
    RGB = np.round(rgb * 255)
    RGB[RGB > 255] = 255
    RGB[RGB < 0] = 0
    lut = np.column_stack((RGB, np.repeat(255, len(r))))
    lut = np.array(lut, 'i')
    return lut


def uv_wheel_lut_table2(points):
    b = np.sum(points, axis=0) / points.shape[0]
    P = np.copy(points)
    for i in range(points.shape[1]):
        P[:, i] -= b[i]
    C = np.einsum('ij,jk->ik', P.T, P)
    eig = np.linalg.eigh(C)
    w = eig[1][:, np.flip(np.argsort(np.abs(eig[0])))]
    P = np.dot(P, w)
    C = (np.max(P, axis=0) + np.min(P, axis=0)) / 2
    P[:, 0] -= C[0]
    P[:, 1] -= C[1]
    P[:, 2] -= C[2]
    P[:, 0] /= np.max(np.abs(P[:, 0]))
    P[:, 1] /= np.max(np.abs(P[:, 1]))
    P[:, 2] /= np.max(np.abs(P[:, 2]))
    r = np.max(np.abs(P[:, [0, 1]]), axis=1)
    v = utilities.normalize(P[:, [0, 1]], axis=1)
    a = np.degrees(np.arccos(v[:, 0]))
    a[v[:, 1] < 0] = 360 - a[v[:, 1] < 0]
    rgb = hsv_to_rgb(np.column_stack((a, r, 0.95 + 0.05 * P[:, 2])))
    RGB = np.round(rgb * 255)
    RGB[RGB > 255] = 255
    RGB[RGB < 0] = 0
    lut = np.column_stack((RGB, np.repeat(255, len(r))))
    return lut


# ------------------------------------------------------------------------------
#                                   Color
# ------------------------------------------------------------------------------

def hsv_to_rgb(hsv):
    H = hsv[:, 0] / 60
    S = hsv[:, 1]
    V = hsv[:, 2]
    C = V * S
    X = C * (1 - np.abs(np.mod(H, 2) - 1))
    RGB = np.zeros((len(H), 3))
    c = H >= 0
    RGB[c, 0] = C[c]
    RGB[c, 1] = X[c]
    c = H > 1
    RGB[c, 0] = X[c]
    RGB[c, 1] = C[c]
    c = H > 2
    RGB[c, 0] = 0
    RGB[c, 1] = C[c]
    RGB[c, 2] = X[c]
    c = H > 3
    RGB[c, 0] = 0
    RGB[c, 1] = X[c]
    RGB[c, 2] = C[c]
    c = H > 4
    RGB[c, 0] = X[c]
    RGB[c, 1] = 0
    RGB[c, 2] = C[c]
    c = np.logical_and(H > 5, H <= 6)
    RGB[c, 0] = C[c]
    RGB[c, 1] = 0
    RGB[c, 2] = X[c]
    m = V - C
    RGB[:, 0] += m
    RGB[:, 1] += m
    RGB[:, 2] += m
    return (RGB)


# ------------------------------------------------------------------------------
#                            Make the ModuleManager
# ------------------------------------------------------------------------------

def make_module(color, opacity, lut_range, lut_expansion,
                reverse_lut, data_range, **kwargs):
    module = ModuleManager()
    if data_range is not None:
        module.scalar_lut_manager.use_default_range = False
        max_data = data_range[1]
        min_data = data_range[0]
        if max_data == min_data:
            max_data += 1
            min_data -= 1

        if type(lut_range) == str:

            if lut_range == '-:0:+':
                abs_data = max(abs(max_data), abs(min_data))
                lut_range = np.array([-abs_data, abs_data])

            elif lut_range == '0:+':
                lut_range = np.array([0, max_data])

            elif lut_range == '-:0':
                lut_range = np.array([min_data, 0])

            elif lut_range == '-:+':
                lut_range = np.array([min_data, max_data])

        module.scalar_lut_manager.data_range = lut_range

        if isinstance(color, np.ndarray):
            module.scalar_lut_manager.number_of_colors = len(color)
            module.scalar_lut_manager.lut.table = color

        elif color == 'bwr_e':
            lut_e = bwr_lut_table(lut_expansion)
            module.scalar_lut_manager.lut.table = lut_e

        elif color == 'rgbm':
            lut_e = rgbm_lut_table(lut_expansion)
            module.scalar_lut_manager.lut.table = lut_e

        elif color == 'rgb_e':
            lut_e = rgb_lut_table(lut_expansion, opacity)
            module.scalar_lut_manager.lut.table = lut_e

        elif color == 'xyz':
            points = kwargs.get('points')
            lut = xyz_lut_table(points)
            module.scalar_lut_manager.number_of_colors = len(points)
            module.scalar_lut_manager.lut.table = lut
            # module.scalar_lut_manager.use_default_range = True

        elif color == 'uv_wheel':
            points = kwargs.get('points')
            lut = uv_wheel_lut_table(points)
            module.scalar_lut_manager.number_of_colors = len(points)
            module.scalar_lut_manager.lut.table = lut

        elif color == 'turbo':
            lut = colorutilities.turbo_lut_table()
            module.scalar_lut_manager.lut.table = lut

        elif type(color) == list or type(color) == tuple:
            lut_c = lut_table(color, opacity)
            module.scalar_lut_manager.lut.table = lut_c

        else:
            try:
                module.scalar_lut_manager.lut_mode = color
                module.scalar_lut_manager.reverse_lut = reverse_lut
            except:
                lut_c = lut_table(color, opacity)
                module.scalar_lut_manager.lut.table = lut_c
    else:
        lut_c = lut_table(color, opacity)
        module.scalar_lut_manager.lut.table = lut_c
        module.scalar_lut_manager.use_default_range = True

    return module


def make_lut_table(module, color, opacity, lut_expansion, reverse_lut, **kwargs):
    if isinstance(color, np.ndarray):
        try:
            # module.scalar_lut_manager.number_of_colors = len(color)
            # module.scalar_lut_manager.lut.table = color
            # module.scalar_lut_manager.lut_mode = color
            # module.scalar_lut_manager.reverse_lut = reverse_lut
            # module.scalar_lut_manager._lut_mode_changed(color)
            # module.scalar_lut_manager._lut_mode_changed(color)
            pass
        except:
            pass

    elif color == 'bwr_e':
        lut_e = bwr_lut_table(lut_expansion)
        module.scalar_lut_manager.lut.table = lut_e

    elif color == 'rgb_e':
        lut_e = rgb_lut_table(lut_expansion)
        module.scalar_lut_manager.lut.table = lut_e

    elif type(color) == list or type(color) == tuple:
        lut_c = lut_table(color, opacity)
        module.scalar_lut_manager.lut.table = lut_c
        # print(dir(module.scalar_lut_manager))

    else:
        lut_c = lut_table(color, opacity)
        module.scalar_lut_manager.lut.table = lut_c


def make_lut_range(lut_range, data_range):
    max_data = data_range[1]
    min_data = data_range[0]
    if max_data == min_data:
        max_data += 1
        min_data -= 1
    if type(lut_range) == str:
        if lut_range == '-:0:+':
            abs_data = max(abs(max_data), abs(min_data))
            lut_range = np.array([-abs_data, abs_data])
        elif lut_range == '0:+':
            lut_range = np.array([0, max_data])
        elif lut_range == '-:0':
            lut_range = np.array([min_data, 0])
        elif lut_range == '-:+':
            lut_range = np.array([min_data, max_data])
    else:
        lut_range = np.array(lut_range)
    return lut_range


# ------------------------------------------------------------------------------
#                   Make the ModuleManager for Vectors
# ------------------------------------------------------------------------------

def make_vector_module(color, opacity, lut_range, lut_expansion, reverse_lut):
    module = ModuleManager()

    if color == 'bwr_e':
        lut_e = bwr_lut_table(lut_expansion)
        module.vector_lut_manager.lut.table = lut_e

    elif color == 'rgb_e':
        lut_e = rgb_lut_table(lut_expansion)
        module.vector_lut_manager.lut.table = lut_e

    elif type(color) == list or type(color) == tuple:
        lut_c = lut_table(color, opacity)
        module.vector_lut_manager.lut.table = lut_c

    else:
        try:
            module.vector_lut_manager.lut_mode = color
            module.vector_lut_manager.reverse_lut = reverse_lut
        except:
            lut_c = lut_table(color, opacity)
            module.vector_lut_manager.lut.table = lut_c

    module.vector_lut_manager.use_default_range = True
    return module


# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------

def lut_library():
    lut_mode = ['bwr_e', 'rgb_e', 'xyz', 'rgbm', 'uv_wheel', 'turbo',
                'Accent', 'Blues', 'Br_bG', 'Bu_gn', 'Bu_pu', 'CMRmap', 'Dark2', 'Gn_bu',
                'Greens', 'Greys', 'Or_rd', 'Oranges', 'PRGn', 'Paired', 'Pastel1', 'Pastel2',
                'Pi_yG', 'Pu_bu', 'Pu_bu_gn', 'Pu_or', 'Pu_rd', 'Purples', 'Rd_bu', 'Rd_gy', 'Rd_pu',
                'Rd_yl_bu', 'Rd_yl_gn', 'Reds', 'Set1', 'Set2', 'Set3', 'Spectral', 'Vega10', 'Vega20',
                'Vega20b', 'Vega20c', 'Wistia', 'Yl_gn', 'Yl_gn_bu', 'Yl_or_br', 'Yl_or_rd', 'afmhot',
                'autumn', 'binary', 'black-white', 'blue-red', 'bone', 'brg', 'bwr', 'cool',
                'coolwarm', 'copper', 'cubehelix', 'file', 'flag', 'gist_earth', 'gist_gray',
                'gist_heat', 'gist_ncar', 'gist_rainbow', 'gist_stern', 'gist_yarg', 'gnuplot',
                'gnuplot2', 'gray', 'hot', 'hsv', 'inferno', 'jet', 'magma', 'nipy_spectral',
                'ocean', 'pink', 'plasma', 'prism', 'rainbow', 'seismic', 'spectral', 'spring',
                'summer', 'terrain', 'viridis', 'winter',
                ]
    return lut_mode


# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------

def check_arguments(**kwargs):
    point_data = kwargs.get('point_data')
    message = '*point_data* argument must be a list or a np.ndarray!'
    if point_data is not None:
        if type(point_data) is not np.ndarray and type(point_data) is not list:
            raise ValueError(message)
    vertex_data = kwargs.get('vertex_data')
    message = '*vertex_data* argument must be a list or a np.ndarray!'
    if vertex_data is not None:
        if type(vertex_data) is not np.ndarray and type(vertex_data) is not list:
            raise ValueError(message)
    face_data = kwargs.get('face_data')
    message = '*face_data* argument must be a list or a np.ndarray!'
    if face_data is not None:
        if type(face_data) is not np.ndarray and type(face_data) is not list:
            raise ValueError(message)
    edge_data = kwargs.get('edge_data')
    message = '*edge_data* argument must be a list or a np.ndarray!'
    if edge_data is not None:
        if type(edge_data) is not np.ndarray and type(edge_data) is not list:
            raise ValueError(message)
    indices = kwargs.get('vertex_indices')
    message = '*vertex_indices* argument must be a list or a np.ndarray!'
    if indices is not None:
        if type(indices) is not np.ndarray and type(indices) is not list:
            raise ValueError(message)
    color = kwargs.get('color')
    message = '*color* argument must be:\n'
    message += ' - an (r,g,b) tuple with [0,255] floats\n'
    message += ' - a color keyword string: '
    colors = color_library()
    message += str(colors.keys()) + '\n'
    message += ' - a palette keyword string: '
    palettes = palette_library()
    message += str(palettes.keys()) + '\n'
    message += ' - a list of (r,g,b) tuples and/or color keyword strings\n'
    message += ' - a color table keyword string: '
    tables = lut_library()
    message += str(tables) + '\n'
    if color is not None:
        if isinstance(color, np.ndarray):
            pass
        else:
            if type(color) is not list:
                color = [color]
            for c in color:
                if type(c) is tuple and len(c) == 3:
                    for val in c:
                        if type(val) is float or type(val) is int:
                            if val >= 0 and val <= 255:
                                pass
                            else:
                                raise ValueError(message)
                        else:
                            raise ValueError(message)
                elif type(c) is str:
                    if c in colors:
                        pass
                    elif c in palettes:
                        pass
                    elif c in tables and len(color) == 1:
                        pass
                    else:
                        raise ValueError(message)
                else:
                    raise ValueError(message)
    opacity = kwargs.get('opacity')
    message = '*opacity* argument must be a [0,1] float!'
    if opacity is not None:
        if type(opacity) is float or type(opacity) is int:
            if opacity >= 0 and opacity <= 1:
                pass
            else:
                raise ValueError(message)
    glyph_type = kwargs.get('glyph_type')
    message = '*glyph_type* argument must be a keyword string: '
    types = ['cube', 'sphere', '3D-arrow', 'cylinder', 'line', 'cone', 'axes']
    message += str(types)
    if glyph_type is not None:
        if type(glyph_type) is str:
            if glyph_type in types:
                pass
            else:
                raise ValueError(message)
        else:
            raise ValueError(message)
    lut_range = kwargs.get('lut_range')
    message = '*lut_range* argument must be a list or tuple of length 2 '
    message += 'with float elements, '
    message += 'a float np.ndarray of shape (2,) or a keyword string: '
    ranges = ['-:+', '-:0:+', '-:0', '0:+']
    message += str(ranges)
    if lut_range is not None:
        if type(lut_range) is tuple or type(lut_range) is list:
            if len(lut_range) == 2:
                v1 = float(lut_range[0])
                v2 = float(lut_range[1])
                if type(v1) is float and type(v2) is float:
                    pass
                else:
                    raise ValueError(message)
            else:
                raise ValueError(message)
        elif type(lut_range) is np.ndarray:
            print(lut_range.shape)
            if len(lut_range.shape) == 1 and lut_range.shape[0] == 2:
                pass
            else:
                raise ValueError(message)
        elif type(lut_range) is str:
            if lut_range in ranges:
                pass
            else:
                raise ValueError(message)
        else:
            raise ValueError(message)
    anchor_mode = kwargs.get('anchor_mode')
    message = '*anchor_mode* argument must be a keyword string: '
    types = ['vertex', 'face', 'edge']
    message += str(types)
    if anchor_mode is not None:
        if anchor_mode in types:
            pass
        else:
            raise ValueError(message)



# !/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

# -----------------------------------------------------------------------------

from tvtk.api import tvtk

from mayavi.core.api import ModuleManager

from traits.api import HasTraits

from wsurf.geolab.plotter.pointsource import Points

from wsurf.geolab.plotter.facesource import Faces

from wsurf.geolab.plotter.edgesource import Edges

from wsurf.geolab.plotter.vectorsource import Vectors

from wsurf.geolab.plotter.polylinesource import Polyline

# ------------------------------------------------------------------------------

'''plotmanager.py: The plot manager class'''

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                   Plot Manager
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class PlotManager(HasTraits):

    def __init__(self, scene_model=None, name='plot_0'):
        HasTraits.__init__(self)

        self.name = name

        self._scene_model = scene_model

        self._sources = {}

        self._widgets = {}

        self.picker_tolerance = 0.003

        self._position = None

        self._plot_callbacks = {}

        self._cross_callbacks = {}

        self.__counter = 0

    # -------------------------------------------------------------------------
    #                                Properties
    # -------------------------------------------------------------------------

    @property
    def type(self):
        return 'Plot_manager'

    @property
    def scene_model(self):
        return self._scene_model

    @scene_model.setter
    def scene_model(self, scene_model):
        self._scene_model = scene_model

    @property
    def scene(self):
        return self._scene_model.mayavi_scene

    @property
    def background(self):
        return self.scene_model.background

    @background.setter
    def background(self, color):
        self.scene_model.background = color

    @property
    def magnification(self):
        return self.scene_model.magnification

    @magnification.setter
    def magnification(self, magnification):
        self.scene_model.magnification = magnification

    @property
    def position(self):
        return self.get_position()

    @position.setter
    def position(self, position):
        if position is not None:
            # self.set_view(position)
            self._position = position

    @property
    def sources(self):
        return self._sources

    @property
    def number_of_frames(self):
        return len(self._record_register)

    @property
    def shooting(self):
        return self.__shooting

    # -------------------------------------------------------------------------
    #                                    view
    # -------------------------------------------------------------------------

    def parallel_projection(self, parallel=True):
        self.scene.scene.camera.parallel_projection = parallel

    def z_view(self):

        self.scene.scene.camera.compute_view_plane_normal()
        self.scene.z_plus_view()

    def reset_view(self):
        self.scene.scene.z_plus_view()

    def interaction_2d(self):
        self.scene.scene.interactor.interactor_style = tvtk.InteractorStyleImage()

    def interaction_locked(self):
        self.scene.scene.interactor.interactor_style = None

    # -------------------------------------------------------------------------
    #                            Pipeline management
    # -------------------------------------------------------------------------

    def add_widget(self, widget, name='widget'):
        self._widgets[name] = widget
        self.scene_model.add_actors(widget)

    def get_widget(self, name='widget'):
        try:
            return self._widgets[name]
        except KeyError:
            return None

    def remove_widgets(self):
        for key in self._widgets:
            try:
                self.scene_model.remove_actor(self._widgets[key])
            except:
                pass
        self._widgets = {}

    def add(self, objects, plot=True, pickable=False, rename=False):
        if not isinstance(objects, list):
            objects = [objects]
        e = self.scene_model.engine
        self.fix_view()
        n = 0
        for obj in objects:
            if rename:
                obj.name = 'obj_' + str(n)
                n += 1
            try:
                self._sources[obj.name].source.remove()
            except:
                pass
            src = obj.source
            self._sources[obj.name] = obj
            if plot:
                e.add_source(src, scene=self.scene_model)
                obj.on_scene = True
            if not pickable:
                self.set_pickable(pick=False, name=obj.name)
        self.apply_view()

    def clear(self):
        self.fix_view()
        for key in self._sources:
            try:
                self._sources[key].source.remove()
                self._sources[key].on_scene = False
            except:
                pass
        self._sources = {}
        self.remove_widgets()
        self.apply_view()
        # self._plot_callbacks = {}
        # self._cross_callbacks = {}

    def hide(self, name=None, **kwargs):
        self.fix_view()
        if name is None:
            for key in self._sources:
                try:
                    self._sources[key].source.remove()
                    self._sources[key].on_scene = False
                except:
                    pass
        else:
            try:
                self._sources[name].source.remove()
                self._sources[name].on_scene = False
            except:
                pass
        self.apply_view()

    def remove(self, names):
        if type(names) is str:
            names = [names]
        self.fix_view()
        for key in names:
            try:
                self._sources[key].source.remove()
                del self._sources[key]
            except:
                pass
        self.apply_view()

    def keep_only(self, keep):
        for key in self._sources:
            if key not in keep:
                try:
                    self._sources[key].remove()
                    self.remove(key)
                except:
                    pass

    def show(self, name):
        e = self.scene_model.engine
        self.fix_view()
        try:
            e.add_source(self._sources[name].source, scene=self.scene_model)
            self._sources[name].on_scene = True
        except:
            pass
        self.apply_view()

    def update(self, **kwargs):
        name = kwargs.pop('name', None)
        self.fix_view()
        if name not in self._sources:  # in video, this skip deleted sources
            return
        self._sources[name].update(**kwargs)
        if not self._sources[name].on_scene:
            e = self.scene_model.engine
            e.add_source(self._sources[name].source, scene=self.scene_model)
            self._sources[name].on_scene = True
        self.apply_view()

    def has_source(self, name):
        if name in self._sources:
            return True
        else:
            return False

    def is_on_scene(self, name):
        if name in self._sources:
            if self._sources[name].on_scene:
                return True
        return False

    def get_source(self, name):
        try:
            return self.sources[name]
        except:
            return None

    # -------------------------------------------------------------------------
    #                                 Output
    # -------------------------------------------------------------------------

    def save(self, name, size=None):
        scene = self.scene_model
        scene.render_window.point_smoothing = True
        scene.render_window.line_smoothing = True
        scene.render_window.polygon_smoothing = True
        scene.render_window.multi_samples = 8
        if size is None:
            scene.save_png(name)
        else:
            scene.save(name, size)

    # -------------------------------------------------------------------------
    #                              View management
    # -------------------------------------------------------------------------

    def disable_render(self):
        try:
            self.scene.scene.disable_render = True
        except:
            pass

    def enable_render(self):
        try:
            self.scene.scene.disable_render = False
        except:
            pass

    def fix_view(self):
        try:
            self.scene.scene.disable_render = True
            cc = self.scene.scene.camera
            self.__orig_pos = cc.position
            self.__orig_fp = cc.focal_point
            self.__orig_view_angle = cc.view_angle
            self.__orig_view_up = cc.view_up
            self.__orig_clipping_range = cc.clipping_range
            self.__parallel_projection = cc.parallel_projection
        except:
            pass

    def apply_view(self):
        try:
            cc = self.scene.scene.camera
            cc.position = self.__orig_pos
            cc.focal_point = self.__orig_fp
            cc.view_angle = self.__orig_view_angle
            cc.view_up = self.__orig_view_up
            cc.clipping_range = self.__orig_clipping_range
            cc.parallel_projection = self.__parallel_projection
            self.scene.scene.disable_render = False
        except:
            pass

    def get_position(self):
        scene = self.scene.scene
        cc = scene.camera
        p = [cc.position[0], cc.position[1], cc.position[2]]
        p.extend([cc.view_up[0], cc.view_up[1], cc.view_up[2]])
        p.extend([cc.focal_point[0], cc.focal_point[1], cc.focal_point[2]])
        p.extend([cc.view_angle])
        p.extend([cc.clipping_range[0], cc.clipping_range[1]])
        p.extend([scene.parallel_projection])
        return p

    def vertical_view(self):
        position = np.array(self.get_position())
        position[3:6] = np.array([0, 0, 1])
        self.set_view(position)

    def camera_rotation(self, degrees):
        center = self.position[6:9]
        position = np.array(self.get_position())
        P = position[0:2]
        F = position[6:8]
        cos = np.cos(np.radians(degrees))
        sin = np.sin(np.radians(degrees))
        V = P - center[0:2]
        Vx = V[0] * cos - V[1] * sin + center[0]
        Vy = V[0] * sin + V[1] * cos + center[1]
        position[0] = Vx
        position[1] = Vy
        V = F - center[0:2]
        Vx = V[0] * cos - V[1] * sin + center[0]
        Vy = V[0] * sin + V[1] * cos + center[1]
        position[6] = Vx
        position[7] = Vy
        position[3:6] = np.array([0, 0, 1])
        self.set_view(position)

    def save_view(self, print_position=True):
        self._position = self.get_position()
        if print_position:
            print('position = ' + str(self._position))

    def set_view(self, position=None):
        cc = self.scene.scene.camera
        p = self._position
        if position is not None:
            p = position
        if p is not None:
            cc.position = np.array([p[0], p[1], p[2]])
            cc.view_up = np.array([p[3], p[4], p[5]])
            cc.focal_point = np.array([p[6], p[7], p[8]])
            cc.view_angle = np.float(p[9])
            cc.clipping_range = np.array([p[10], p[11]])
            try:
                cc.parallel_projection = p[12]
            except:
                pass
            self.scene.render()

    def render(self):
        self.scene.scene._renwin.render()

    # -------------------------------------------------------------------------
    #                            Get from pipeline
    # -------------------------------------------------------------------------

    def get_actor(self, name):
        stop = False
        try:
            obj = self._sources[name].source
        except KeyError:
            return
        while not stop:
            obj = obj.children[0]
            if isinstance(obj, ModuleManager):
                stop = True
        actor = obj.children[0].actor
        return actor

    # -------------------------------------------------------------------------
    #                                 Picker
    # -------------------------------------------------------------------------

    def set_pickable(self, pick=True, name=None):
        from mayavi.core.api import ModuleManager
        if name is None:
            for key in self._sources:
                obj = self._sources[key].source
                if pick:
                    pick = 1
                else:
                    pick = 0
                stop = False
                while not stop:
                    obj = obj.children[0]
                    if isinstance(obj, ModuleManager):
                        stop = True
                actor = obj.children[0].actor.actors[0]
                actor.pickable = pick
        else:
            try:
                obj = self._sources[name].source
            except:
                return
            if pick:
                pick = 1
            else:
                pick = 0
            stop = False
            while not stop:
                obj = obj.children[0]
                if isinstance(obj, ModuleManager):
                    stop = True
            actor = obj.children[0].actor.actors[0]
            actor.pickable = pick

    def picker_callback(self, routine, mode='point', name=None, add=False):
        if name is not None:
            self.set_pickable(False)
            self.set_pickable(True, name)
        else:
            self.set_pickable(True)

        def picker_callback(picker):
            pick_id = picker.cell_id
            if mode == 'point':
                pick_id = picker.cell_id
            routine(pick_id)

        # s = self.scene_model.engine.current_scene
        s = self.scene
        p = s._mouse_pick_dispatcher.callbacks
        if add:
            s.on_mouse_pick(picker_callback, 'cell')
        elif len(p) == 0:
            s.on_mouse_pick(picker_callback, 'cell')
        else:
            p[0] = (picker_callback, 'cell', 'Left')
        a = s._mouse_pick_dispatcher._active_pickers['cell']
        a.tolerance = self.picker_tolerance

    def picker_off(self):
        # scene = self.scene_model.engine.current_scene
        p = self.scene._mouse_pick_dispatcher.callbacks

        def picker_callback(picker):
            return

        for i in range(len(p)):
            p[i] = (picker_callback, 'cell', 'Left')

    def iterate(self, function, times=1):
        for i in range(times):
            function()
            self.scene.scene._renwin.render()

    # -------------------------------------------------------------------------
    #                                 Plot
    # -------------------------------------------------------------------------

    def start_plot(self):
        self.disable_render()

    def end_plot(self):
        self.enable_render()

    def core_plot(self):
        for key in self._plot_callbacks:
            callback = self._plot_callbacks[key]
            try:
                callback()
            except:
                pass

    def cross_plot(self, main):
        if main:
            for key in self._cross_callbacks:
                callback = self._cross_callbacks[key]
                try:
                    callback(main=False)
                except:
                    pass

    def update_plot(self, main=True):
        self.start_plot()
        self.core_plot()
        # self.cross_plot(main)
        self.end_plot()
        # print('plm_update_plot')

    def add_plot_callback(self, callback, name=None):
        if callable(callback):
            if name is None:
                name = 'plot_callback_{}'.format(self.__counter)
                self.__counter += 1
            self._plot_callbacks[name] = callback

    def add_cross_callback(self, callback, name=None, **kwargs):
        if callable(callback):
            if name is None:
                name = 'cross_callback_{}'.format(self.__counter)
                self.__counter += 1
            self._cross_callbacks[name] = callback

    def remove_plot_callback(self, name):
        try:
            del (self._plot_callbacks[name])
        except:
            pass

    # -------------------------------------------------------------------------
    #                              Plot Functions
    # -------------------------------------------------------------------------

    def plot_points(self, points, **kwargs):
        name = kwargs.pop('name', 'points')
        kwargs['name'] = name
        if name in self._sources:
            kwargs['points'] = points
            self.update(**kwargs)
        else:
            P = Points(points, **kwargs)
            self.add(P)

    def plot_polyline(self, polyline, **kwargs):
        name = kwargs.pop('name', 'polyline')
        kwargs['name'] = name
        if name in self._sources:
            kwargs['polyline'] = polyline
            self.update(**kwargs)
        else:
            P = Polyline(polyline, **kwargs)
            self.add(P)

    def plot_mesh_faces(self, mesh, **kwargs):
        name = kwargs.pop('name', 'mesh-faces')
        kwargs['name'] = name
        if name in self._sources:
            kwargs['mesh'] = mesh
            self.update(**kwargs)
        else:
            P = Faces(mesh, **kwargs)
            self.add(P)

    def plot_mesh_edges(self, mesh, **kwargs):
        name = kwargs.pop('name', 'mesh-edges')
        kwargs['name'] = name
        if name in self._sources:
            kwargs['mesh'] = mesh
            self.update(**kwargs)
        else:
            P = Edges(mesh, **kwargs)
            self.add(P)

    def plot_vectors(self, vectors, **kwargs):
        name = kwargs.pop('name', 'vectors')
        kwargs['name'] = name
        if name in self._sources:
            kwargs['vectors'] = vectors
            self.update(**kwargs)
        else:
            kwargs['name'] = name
            P = Vectors(vectors, **kwargs)
            self.add(P)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from scipy import sparse

from scipy import spatial

# ------------------------------------------------------------------------------

from wsurf.geolab import utilities

# -----------------------------------------------------------------------------

__author__ = 'Davide Pellis'


# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class BsplineSurface(object):

    def __init__(self, file_name=None, u_degree=3, v_degree=3,
                 control_points=None, u_knot_vector=None, v_knot_vector=None):

        self.name = 'bspline-surface'

        self._u_degree = u_degree

        self._v_degree = v_degree

        self._control_points = control_points

        self._control_points_index = None

        self._u_knot_vector = u_knot_vector

        self._v_knot_vector = v_knot_vector

        self._u_sampling = 100

        self._v_sampling = 100

        self._u_range = None

        self._v_range = None

        self._u = None

        self._v = None

        self.plot_mode = 'surface'

        self._sealed = [False, False]

        self._kdtree = None

        if file_name is not None:
            self.read_obj_file(file_name)

        print(self)

    # --------------------------------------------------------------------------

    def __str__(self):
        n = 'B-spline surface: '
        info = 'degree = ({},{}), number of control points = '
        out = n + info.format(self._u_degree, self._v_degree)
        degree = '({}'
        U = self.Cu
        V = self.Cv
        if self._sealed[0]:
            U -= self.u_degree
            degree += 'c'
        degree += ',{}'
        if self._sealed[1]:
            V -= self.v_degree
            degree += 'c'
        degree += ')'
        return out + degree.format(U, V)

    # --------------------------------------------------------------------------
    # --------------------------------------------------------------------------

    @property
    def type(self):
        return 'Bspline_Surface'

    @property
    def u_sealed(self):
        return self._sealed[0]

    @property
    def v_sealed(self):
        return self._sealed[1]

    @property
    def number_of_control_points(self):
        U = self.Cu
        V = self.Cv
        if self._sealed[0]:
            U -= self.u_degree
        if self._sealed[1]:
            V -= self.v_degree
        return (U, V)

    @property
    def C(self):
        return len(self._control_points)

    @property
    def Cu(self):
        return len(self._u_knot_vector) - self._u_degree - 1

    @property
    def Cv(self):
        return len(self._v_knot_vector) - self._v_degree - 1

    @property
    def S(self):
        return self.u_sampling * self.v_sampling

    @property
    def u_degree(self):
        return self._u_degree

    @property
    def v_degree(self):
        return self._v_degree

    @property
    def u(self):
        if self._u is None:
            self._u = np.linspace(self._u_range[0], self._u_range[1],
                                  self.u_sampling)
        return self._u

    @u.setter
    def u(self, u):
        if type(u) == int or type(u) == float:
            u = np.array(u)
        self._u = u
        self._u_sampling = len(u)

    @property
    def v(self):
        if self._v is None:
            self._v = np.linspace(self._v_range[0], self._v_range[1],
                                  self.v_sampling)
        return self._v

    @v.setter
    def v(self, v):
        if type(v) == int or type(v) == float:
            v = np.array(v)
        self._v = v
        self._v_sampling = len(v)

    @property
    def u_sampling(self):
        return self._u_sampling

    @u_sampling.setter
    def u_sampling(self, nu):
        self._u_sampling = int(nu)
        self._u = None

    @property
    def v_sampling(self):
        return self._v_sampling

    @v_sampling.setter
    def v_sampling(self, nv):
        self._v_sampling = int(nv)
        self._v = None

    @property
    def control_points(self):
        return self._control_points

    @control_points.setter
    def control_points(self, control_points):
        self._control_points = control_points

    @property
    def d(self):
        return self.control_points.shape[1]

    @property
    def V(self):
        if self.plot_mode == 'surface':
            return self.S
        else:
            return self.C

    @property
    def E(self):
        if self.plot_mode == 'surface':
            E = ((self.u_sampling - 1) * self.v_sampling +
                 (self.v_sampling - 1) * self.u_sampling)
            return E
        else:
            return (self.Cu - 1) * self.Cv + (self.Cv - 1) * self.Cu

    @property
    def vertices(self):
        if self.plot_mode == 'surface':
            return self.points()
        else:
            return self.control_points

    @property
    def degree(self):
        return self._u_degree, self._v_degree

    @property
    def sampling(self):
        return (self.u_sampling, self.v_sampling)

    @sampling.setter
    def sampling(self, sampling):
        self.u_sampling = sampling[0]
        self.v_sampling = sampling[1]

    # --------------------------------------------------------------------------
    #                             Reading - writing
    # --------------------------------------------------------------------------

    def read_obj_file(self, file_name):
        file_name = str(file_name)
        self.name = file_name.split('.')[0]
        obj_file = open(file_name, encoding='utf-8')
        vertices_list = []
        read = 0
        control = False
        for l in obj_file:
            splited_line = l.split(' ')
            if splited_line[0] == 'v':
                split_x = splited_line[1].split('\n')
                x = float(split_x[0])
                split_y = splited_line[2].split('\n')
                y = float(split_y[0])
                split_z = splited_line[3].split('\n')
                try:
                    z = float(split_z[0])
                except ValueError:
                    print('WARNING: disable line wrap when saving .obj')
                vertices_list.append([x, y, z])
            elif splited_line[0] == 'deg':
                if len(splited_line) == 3:
                    self._u_degree = int(splited_line[1])
                    self._v_degree = int(splited_line[2])
                    read += 1
            if read == 1:
                if splited_line[0] == 'surf':
                    self._u_range = np.array((float(splited_line[1]),
                                              float(splited_line[2])))
                    self._v_range = np.array((float(splited_line[3]),
                                              float(splited_line[4])))
                    indices = []
                    for i in range(len(splited_line) - 5):
                        indices.append(int(splited_line[i + 5]) - 1)
                if splited_line[0] == 'parm':
                    if splited_line[1] == 'u':
                        u_knot = []
                        for i in range(len(splited_line) - 2):
                            u_knot.append(float(splited_line[i + 2]))
                    elif splited_line[1] == 'v':
                        v_knot = []
                        for i in range(len(splited_line) - 2):
                            v_knot.append(float(splited_line[i + 2]))
                        read += 1
                        control = True
        if not control:
            raise ValueError('unable to read the file')
        self._control_points = np.array(vertices_list)
        self._control_points_index = np.array(indices, 'i')
        self._u_knot_vector = np.array(u_knot)
        self._v_knot_vector = np.array(v_knot)
        self._seal()

    def make_obj_file(self, file_name, overwrite=False):
        """Save the bspline surface as OBJ file.

        Parameters
        ----------
        file_name : str
            The path of the OBJ file to be created, without extension.
        overwrite : bool
            If `False` (default), when the path already exists, a sequential
            number is added to file_name. If `True`, existing paths are
            overwritten.

        Returns
        -------
        str
            The path of the newly generated OBJ file (without extension).
        """
        path = utilities.make_filepath(file_name, 'obj', overwrite)
        obj = open(path, 'w')
        line = 'o {}\n'.format(file_name)
        obj.write(line)
        for c in range(self.C):
            ci = self.control_points[c]
            line = 'v {} {} {}\n'.format(ci[0], ci[1], ci[2])
            obj.write(line)
        line = 'cstype bspline\n'
        obj.write(line)
        line = 'deg {} {}\n'.format(self._u_degree, self._v_degree)
        obj.write(line)
        line = 'surf {} {} {} {}'.format(self._u_range[0], self._u_range[1],
                                         self._v_range[0], self._v_range[1])
        for c in self._control_points_index:
            line += ' {}'.format(c + 1)
        line += '\n'
        obj.write(line)
        line = 'parm u'
        for u in self._u_knot_vector:
            line += ' {}'.format(u)
        line += '\n'
        obj.write(line)
        line = 'parm v'
        for v in self._v_knot_vector:
            line += ' {}'.format(v)
        line += '\n'
        line += 'end'
        obj.write(line)
        obj.close()
        out = 'Obj saved in {}'.format(path)
        print(out)
        return path.split('.')[0]

    # --------------------------------------------------------------------------
    #                                  Seal
    # --------------------------------------------------------------------------

    def _seal_check(self):
        c = self._control_points_index
        c = np.reshape(self._control_points_index, (self.Cu, self.Cv), 'F')
        if len(np.unique(c[:, 0])) < len(c[:, 0]):
            self._sealed[0] = True
        if len(np.unique(c[0, :])) < len(c[0, :]):
            self._sealed[1] = True

    def _seal(self, epsilon=1e-5):
        v = self._control_points_index
        if len(np.unique(v)) < len(v):
            self._seal_check()
        else:
            u_seal = 0
            v_seal = 0
            ud = self.u_degree
            vd = self.v_degree
            for i in range(self.d):
                G = np.reshape(self.control_points[v, i], (self.Cu, self.Cv), 'F')
                u_seal += np.sum(np.abs(G[:ud, :] - G[-ud:, :]))
                v_seal += np.sum(np.abs(G[:, :vd] - G[:, -vd:]))
            ind = np.copy(np.reshape(v, (self.Cu, self.Cv), 'F'))
            c_map = np.reshape(v, (self.Cu, self.Cv), 'F')
            if u_seal < epsilon:
                ind[-ud:, :] = ind[:ud, :]
                c_map = c_map[:-ud, :]
                ind[:, np.arange(self.Cv)] -= ud * np.arange(self.Cv)
                self._sealed[0] = True
            if v_seal < epsilon:
                ind[:, -vd:] = ind[:, :vd]
                c_map = c_map[:, :-vd]
                # may have problems with non arange indices v
                ind -= np.min(ind[:, :vd])
                self._sealed[1] = True
            if self._sealed[0] or self._sealed[1]:
                self._control_points_index = ind.flatten('F')
                self._control_points = self._control_points[c_map.flatten('F')]

    # -------------------------------------------------------------------------
    #                                Evaluation
    # -------------------------------------------------------------------------

    def basis_functions(self):
        Nu = utilities.bspline_basis_functions(self.u, self.u_degree,
                                               self._u_knot_vector)
        Nv = utilities.bspline_basis_functions(self.v, self.v_degree,
                                               self._v_knot_vector)
        return Nu, Nv

    def derivative_basis_functions(self, d=1):
        Du = utilities.bspline_derivative_basis_functions(self.u,
                                                          self.u_degree,
                                                          self._u_knot_vector,
                                                          d=d)
        Dv = utilities.bspline_derivative_basis_functions(self.v,
                                                          self.v_degree,
                                                          self._v_knot_vector,
                                                          d=d)
        return Du, Dv

    def points(self):
        Nu, Nv = self.basis_functions()
        points = []
        for i in range(self.d):
            v = self._control_points_index
            G = np.reshape(self.control_points[v, i], (self.Cu, self.Cv), 'F')
            Pi = np.einsum('ij,ui,vj->uv', G, Nu, Nv)
            points.append(np.reshape(Pi, self.S, 'F'))
        return np.column_stack(points)

    def u_lines(self):
        u = np.unique(self._u_knot_vector)
        Nu = utilities.bspline_basis_functions(u, self.u_degree,
                                               self._u_knot_vector)
        Nv = utilities.bspline_basis_functions(self.v, self.v_degree,
                                               self._v_knot_vector)
        uc = []
        for i in range(self.d):
            v = self._control_points_index
            G = np.reshape(self.control_points[v, i], (self.Cu, self.Cv), 'F')
            P = np.einsum('ij,ui,vj->uv', G, Nu, Nv)
            uc.append(P)
        uc = np.dstack(uc)
        return uc

    def v_lines(self):
        v = np.unique(self._v_knot_vector)
        Nu = utilities.bspline_basis_functions(self.u, self.u_degree,
                                               self._u_knot_vector)
        Nv = utilities.bspline_basis_functions(v, self.v_degree,
                                               self._v_knot_vector)
        vc = []
        for i in range(self.d):
            v = self._control_points_index
            G = np.reshape(self.control_points[v, i], (self.Cu, self.Cv), 'F')
            P = np.einsum('ij,ui,vj->uv', G, Nu, Nv)
            vc.append(P.T)
        vc = np.dstack(vc)
        return vc

    def uv_points(self):
        u = np.tile(self.u, self.v_sampling)
        v = np.repeat(self.v, self.u_sampling)
        return np.column_stack((u, v))

    def first_derivatives(self):
        Nu, Nv = self.basis_functions()
        Nuu, Nvv = self.derivative_basis_functions(d=1)
        Du = np.empty((self.S, 0))
        Dv = np.empty((self.S, 0))
        for i in range(self.d):
            v = self._control_points_index
            G = np.reshape(self.control_points[v, i], (self.Cu, self.Cv), 'F')
            Ui = np.einsum('ij,ui,vj->uv', G, Nuu, Nv)
            Vi = np.einsum('ij,ui,vj->uv', G, Nu, Nvv)
            Du = np.column_stack((Du, np.reshape(Ui, self.S, 'F')))
            Dv = np.column_stack((Dv, np.reshape(Vi, self.S, 'F')))
        return Du, Dv

    def second_derivatives(self, return_first=True):
        Nu, Nv = self.basis_functions()
        Nu_u, Nv_v = self.derivative_basis_functions(d=1)
        Nu_uu, Nv_vv = self.derivative_basis_functions(d=2)
        if return_first:
            Du = np.empty((self.S, 0))
            Dv = np.empty((self.S, 0))
        Duu = np.empty((self.S, 0))
        Duv = np.empty((self.S, 0))
        Dvv = np.empty((self.S, 0))
        for i in range(self.d):
            v = self._control_points_index
            G = np.reshape(self.control_points[v, i], (self.Cu, self.Cv), 'F')
            if return_first:
                Du_i = np.einsum('ij,ui,vj->uv', G, Nu_u, Nv)
                Dv_i = np.einsum('ij,ui,vj->uv', G, Nu, Nv_v)
                Du = np.column_stack((Du, np.reshape(Du_i, self.S, 'F')))
                Dv = np.column_stack((Dv, np.reshape(Dv_i, self.S, 'F')))
            Duu_i = np.einsum('ij,ui,vj->uv', G, Nu_uu, Nv)
            Duv_i = np.einsum('ij,ui,vj->uv', G, Nu_u, Nv_v)
            Dvv_i = np.einsum('ij,ui,vj->uv', G, Nu, Nv_vv)
            Duu = np.column_stack((Duu, np.reshape(Duu_i, self.S, 'F')))
            Duv = np.column_stack((Duv, np.reshape(Duv_i, self.S, 'F')))
            Dvv = np.column_stack((Dvv, np.reshape(Dvv_i, self.S, 'F')))
        if return_first:
            return Du, Dv, Duu, Duv, Dvv
        else:
            return Duu, Duv, Dvv

    def first_fundamental_form(self, Du=None, Dv=None):
        if Du is None or Dv is None:
            Du, Dv = self.first_derivatives()
        E = np.einsum('ij,ij->i', Du, Du)
        F = np.einsum('ij,ij->i', Du, Dv)
        G = np.einsum('ij,ij->i', Dv, Dv)
        return E, F, G

    def fundamental_forms(self, return_curvatures=True):
        Du, Dv, Duu, Duv, Dvv = self.second_derivatives()
        E = np.einsum('ij,ij->i', Du, Du)
        F = np.einsum('ij,ij->i', Du, Dv)
        G = np.einsum('ij,ij->i', Dv, Dv)
        n = utilities.normalize(np.cross(Du, Dv))
        L = np.einsum('ij,ij->i', n, Duu)
        M = np.einsum('ij,ij->i', n, Duv)
        N = np.einsum('ij,ij->i', n, Dvv)
        if return_curvatures:
            K = (L * N - M ** 2) / (E * G - F ** 2)
            H = (L * G - 2 * M * F + N * E) / (2 * (E * G - F ** 2))
            return E, F, G, L, M, N, K, H
        else:
            return E, F, G, L, M, N

    def curvatures(self):
        E, F, G, L, M, N, K, H = self.fundamental_forms()
        k1 = H + np.abs(H ** 2 - K) ** .5
        k2 = H - np.abs(H ** 2 - K) ** .5
        return K, H, k1, k2

    def closest_point(self, points):
        kdtree = spatial.cKDTree(self.points())
        return kdtree.query(points)[1]

    def normals(self):
        Du, Dv = self.first_derivatives()
        N = np.cross(Du, Dv)
        N = utilities.normalize(N)
        return N

    # --------------------------------------------------------------------------
    #                                 Iterators
    # --------------------------------------------------------------------------

    def point_basis_functions_iterators(self):
        c = self._control_points_index
        i = np.tile(np.arange(self.Cu), self.Cv)
        j = np.repeat(np.arange(self.Cv), self.Cu)
        return c, i, j

    def uv_sampling_iterators(self):
        """
        V[s] = P[c] Nu[u,i] Nv[v,j]
        """
        c, i, j = self.point_basis_functions_iterators()
        s = np.repeat(np.arange(self.S), len(c))
        u = np.tile(np.arange(self.u_sampling), self.v_sampling)
        v = np.repeat(np.arange(self.v_sampling), self.u_sampling)
        u = np.repeat(u, self.Cu * self.Cv)
        v = np.repeat(v, self.Cu * self.Cv)
        c = np.tile(c, self.S)
        i = np.tile(i, self.S)
        j = np.tile(j, self.S)
        return s, u, v, c, i, j

    def uv_central_differences_iterators(self, return_uv_delta=False,
                                         include_boundary=True):
        U = self.u_sampling
        V = self.v_sampling
        M = np.arange(U * V)
        M = M.reshape((U, V), order='F')
        if include_boundary:
            M = np.column_stack((M[:, 0], M, M[:, -1]))
            M = np.vstack((M[0, :], M, M[-1, :]))
        up = M[2:, 1:-1].flatten('F')
        um = M[:-2, 1:-1].flatten('F')
        vp = M[1:-1, 2:].flatten('F')
        vm = M[1:-1, :-2].flatten('F')
        p = M[1:-1, 1:-1].flatten('F')
        if return_uv_delta:
            du = np.hstack((self.u[1] - self.u[0], self.u[2:] - self.u[:-2],
                            self.u[-1] - self.u[-2]))
            dv = np.hstack((self.v[1] - self.v[0], self.v[2:] - self.v[:-2],
                            self.v[-1] - self.v[-2]))
            du = np.tile(du, V)
            dv = np.tile(dv, U)
            return p, up, um, vp, vm, du, dv
        else:
            return p, up, um, vp, vm

    # def uv_second_differences_iterators(self):
    #     U = self.u_sampling
    #     V = self.v_sampling
    #     M = np.arange(U*V)
    #     M = M.reshape((U,V), order='F')
    #     s = M[2:-2,2:-2].flatten('F')
    #     up = M[3:-1,2:-2].flatten('F')
    #     um = M[1:-3,2:-2].flatten('F')
    #     vp = M[2:-2,3:-1].flatten('F')
    #     vm = M[2:-2,1:-3].flatten('F')
    #     upp = M[4:,2:-2].flatten('F')
    #     umm = M[:-4,2:-2].flatten('F')
    #     vpp = M[2:-2,4:].flatten('F')
    #     vmm = M[2:-2,:-4].flatten('F')
    #     return s, upp, up, um, umm, vpp, vp, vm, vmm

    def uv_central_differences(self, f, return_uv_delta=False,
                               include_boundary=True):
        p, up, um, vp, vm, du, dv = self.uv_central_differences_iterators(True)
        Du = (f[up] - f[um]) / du
        Dv = (f[vp] - f[vm]) / dv
        if return_uv_delta:
            return Du, Dv, du, dv
        else:
            return Du, Dv

    # --------------------------------------------------------------------------
    #                                   Plot
    # --------------------------------------------------------------------------

    def cell_arrays(self):
        if self.plot_mode == 'surface':
            V = self.v_sampling
            U = self.u_sampling
        else:
            V = self.Cv
            U = self.Cu
        M = np.arange(V * U, dtype=np.int)
        M = np.reshape(M, (V, U))
        Q = np.zeros((V - 1, U - 1, 4), dtype=np.int)
        Q[:, :, 3] = M[:M.shape[0] - 1, :M.shape[1] - 1]
        Q[:, :, 2] = M[:M.shape[0] - 1, 1:]
        Q[:, :, 1] = M[1:, 1:]
        Q[:, :, 0] = M[1:, :M.shape[1] - 1]
        Q = np.reshape(Q, ((V - 1) * (U - 1), 4), 'C')
        Q = np.column_stack((4 * np.ones((V - 1) * (U - 1)), Q))
        Q = np.reshape(Q, Q.shape[0] * Q.shape[1], 'C')
        self.F = (V - 1) * (U - 1)
        types = np.repeat(1, self.F)
        return Q, types

    def edge_vertices(self, mode='control'):
        if mode == 'surface':
            V = self.v_sampling
            U = self.u_sampling
            M = np.arange(V * U, dtype=np.int)
        else:
            V = self.Cv
            U = self.Cu
            M = self._control_points_index
        M = np.reshape(M, (V, U))
        v1 = np.reshape(M[:, :-1], (V * (U - 1)), 'F')
        v2 = np.reshape(M[:, 1:], (V * (U - 1)), 'F')
        v1 = np.hstack((v1, np.reshape(M.T[:, :-1], ((V - 1) * U), 'F')))
        v2 = np.hstack((v2, np.reshape(M.T[:, 1:], ((V - 1) * U), 'F')))
        # self.E = len(v1)
        return v1, v2

    def face_diagonals(self):
        V = self.Cv
        U = self.Cu
        M = self._control_points_index
        M = np.reshape(M, (V, U))
        v1 = M[:-1,:-1].flatten('F')
        v2 = M[1:,1:].flatten('F')
        v3 = M[:-1, 1:].flatten('F')
        v4 = M[1:, :-1].flatten('F')
        d1 = np.hstack((v1, v3))
        d2 = np.hstack((v2, v4))
        return d1, d2

    def inner_vertices(self):
        V = self.v_sampling
        U = self.u_sampling
        M = np.arange(V * U, dtype=np.int)
        M = np.reshape(M, (V, U))
        inner = M[1:-1, 1:-1]
        return inner.flatten()

    # --------------------------------------------------------------------------
    #                            Boundaries
    # --------------------------------------------------------------------------

    def boundary_control_points(self):
        c = self._control_points_index
        c = np.reshape(self._control_points_index, (self.Cu, self.Cv), 'F')
        boundary = np.array([], 'i')
        if not self._sealed[0]:
            boundary = np.hstack((boundary, c[0, :], c[-1, :]))
        if not self._sealed[1]:
            boundary = np.hstack((boundary, c[:, 0], c[:, -1]))
        return np.unique(boundary)

    def boundary_control_curves(self):
        c = self._control_points_index
        c = np.reshape(self._control_points_index, (self.Cu, self.Cv), 'F')
        curves = []
        if not self._sealed[0]:
            curves.append(c[0, :])
            curves.append(c[-1, :])
        if not self._sealed[1]:
            curves.append(c[:, 0])
            curves.append(c[:, -1])
        return curves

    def control_points_curves_iterators(self):
        c = self._control_points_index
        c = np.reshape(self._control_points_index, (self.Cu, self.Cv), 'F')
        cu = (c[1:-1, :]).flatten('F')
        cu, iu = np.unique(cu, return_index=True)
        up = ((c[2:, :]).flatten('F'))[iu]
        um = ((c[:-2, :]).flatten('F'))[iu]
        cv = (c[:, 1:-1]).flatten('F')
        cv, iv = np.unique(cv, return_index=True)
        vp = ((c[:, 2:]).flatten('F'))[iv]
        vm = ((c[:, :-2]).flatten('F'))[iv]
        c = np.hstack((cu, cv))
        cp = np.hstack((up, vp))
        cm = np.hstack((um, vm))
        return c, cp, cm

    # --------------------------------------------------------------------------
    #                             Utilities
    # --------------------------------------------------------------------------

    def scale(self, factor):
        self._control_points *= factor
        self._u_knot_vector *= factor
        self._v_knot_vector *= factor
        self._u_range *= factor
        self._v_range *= factor
        self._u = None
        self._v = None

    def uv_faces(self):
        V = self.v_sampling
        U = self.u_sampling
        M = np.arange(V * U, dtype=np.int)
        M = np.reshape(M, (V, U))
        Q = np.zeros((V - 1, U - 1, 4), dtype=np.int)
        Q[:, :, 3] = M[:M.shape[0] - 1, :M.shape[1] - 1]
        Q[:, :, 2] = M[:M.shape[0] - 1, 1:]
        Q[:, :, 1] = M[1:, 1:]
        Q[:, :, 0] = M[1:, :M.shape[1] - 1]
        Q = np.reshape(Q, ((V - 1) * (U - 1), 4), 'C')
        return Q

    # --------------------------------------------------------------------------
    #                              Rebuild
    # --------------------------------------------------------------------------

    def rebuild(self, u_degree=None, v_degree=None, u_points=None,
                v_points=None):
        if u_degree is None:
            u_degree = self.u_degree
        if v_degree is None:
            v_degree = self.v_degree
        if u_points is None:
            u_points = self.Cu
        elif self._sealed[0]:
            u_points = max(u_points, u_degree)
            u_points += u_degree
        else:
            u_points = max(u_points, u_degree + 1)
        if v_points is None:
            v_points = self.Cv
        elif self._sealed[1]:
            v_points = max(v_points, v_degree)
            v_points += v_degree
        else:
            v_points = max(v_points, v_degree + 1)
        ind = np.reshape(np.arange(u_points * v_points), (u_points, v_points), 'F')
        u_knot = np.linspace(0, 1, u_points - u_degree + 1)
        if not self._sealed[0]:
            u_knot = np.hstack((np.repeat(0, u_degree), u_knot,
                                np.repeat(1, u_degree)))
        else:
            delta = u_knot[1] - u_knot[0]
            i = np.arange(u_degree) + 1
            u_knot = np.hstack((-np.flip(i) * delta, u_knot, 1 + i * delta))
            ind[-u_degree:, :] = ind[:u_degree, :]
            ind[:, np.arange(v_points)] -= u_degree * np.arange(v_points)
        v_knot = np.linspace(0, 1, v_points - v_degree + 1)
        if not self._sealed[1]:
            v_knot = np.hstack((np.repeat(0, v_degree), v_knot,
                                np.repeat(1, v_degree)))
        else:
            delta = v_knot[1] - v_knot[0]
            i = np.arange(v_degree) + 1
            v_knot = np.hstack((-np.flip(i) * delta, v_knot, 1 + i * delta))
            ind[:, -v_degree:] = ind[:, :v_degree]
            ind -= np.min(ind[:, :v_degree])
        P0 = self.points()
        ut = np.linspace(0, 1, self.u_sampling)
        vt = np.linspace(0, 1, self.v_sampling)
        Nu = utilities.bspline_basis_functions(ut, u_degree, u_knot)
        Nv = utilities.bspline_basis_functions(vt, v_degree, v_knot)
        self._control_points_index = ind.flatten('F')
        self._u_knot_vector = u_knot
        self._v_knot_vector = v_knot
        self._u_degree = u_degree
        self._v_degree = v_degree
        self._u = None
        self._v = None
        self._u_range = np.array([0, 1])
        self._v_range = np.array([0, 1])
        s, u, v, c, cu, cv = self.uv_sampling_iterators()
        Cn = np.max(ind) + 1
        X = []
        H = sparse.coo_matrix((Nu[u, cu] * Nv[v, cv], (s, c)), shape=(self.S, Cn))
        H = H.toarray()
        A = np.dot(H.T, H)
        for d in range(self.d):
            Xd = np.linalg.solve(A, np.dot(H.T, P0[:, d]))
            X.append(Xd)
        P = np.column_stack(X)
        self._control_points = P
        self.relax_knot_vector()
        print(self)

    def relax_knot_vector(self):
        Du, Dv = self.first_derivatives()
        fu = np.mean(np.linalg.norm(Du, axis=1))
        fv = np.mean(np.linalg.norm(Dv, axis=1))
        self._u_knot_vector = self._u_knot_vector * fu
        self._u_range = self._u_range * fu
        self._v_knot_vector = self._v_knot_vector * fv
        self._v_range = self._v_range * fv
        self._u = None
        self._v = None


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

if __name__ == '__main__':
    import geolab as geo

    sp = geo.samples.bspline_patch()

    sp.sampling = (100, 40)

    PP = sp.control_points_curves_iterators()

    sp.rebuild(v_points=3, u_points=3, u_degree=3, v_degree=3)

    plot = []

    E, F, G, L, M, N, K, H = sp.fundamental_forms()

    plot.append(geo.plotter.Faces(sp, vertex_data=H, color='blue-red'))

    sp.plot_mode = 'control'

    plot.append(geo.plotter.Edges(sp, color='k'))

    geo.plotter.view(plot)



#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from tvtk.api import tvtk

from mayavi.sources.vtk_data_source import VTKDataSource

from mayavi.modules.vectors import Vectors as vtk_vectors

from mayavi.modules.glyph import Glyph

import numpy as np

# -----------------------------------------------------------------------------

from wsurf.geolab.plotter import plotutilities

# -----------------------------------------------------------------------------

'''vectorsource.py: The vectors plot source class'''

__author__ = 'Davide Pellis'


#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#                                   Vectors
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------


class Vectors:

    def __init__(self, vectors, **kwargs):
        plotutilities.check_arguments(**kwargs)

        self._vectors        = vectors

        self._geometry       = None

        self._anchor         = kwargs.get('anchor',    np.zeros(vectors.shape))

        self._scalars        = kwargs.get('scalars',        None)

        self._anchor_mode    = kwargs.get('anchor_mode',    'vertex')

        self._color          = kwargs.get('color',          'blue-red')

        self._opacity        = kwargs.get('opacity',        1)

        self._line_width     = kwargs.get('line_width',     1)

        self._glyph_type     = kwargs.get('glyph_type',     '3D-arrow')

        self._position       = kwargs.get('position',       'head')

        self._scaling        = kwargs.get('scaling',        True)

        self._scale_factor   = kwargs.get('scale_factor',    1)

        self._radius         = kwargs.get('radius',         None)

        self._tip_radius     = kwargs.get('tip_radius',     None)

        self._tip_length     = kwargs.get('tip_length',     None)

        self._shaft_radius   = kwargs.get('shaft_radius',   None)

        self._height         = kwargs.get('height',         None)

        self._resolution     = kwargs.get('resolution',     20)

        self._lut_range      = kwargs.get('lut_range',      '-:+')

        self._lut_expansion  = kwargs.get('lut_expansion',  2)

        self._reverse_lut    = kwargs.get('reverse_lut',    False)

        self._shading        = kwargs.get('shading',        True)

        self._glossy         = kwargs.get('glossy',         0.3)

        self.name            = kwargs.get('name',           'vectors')

    #--------------------------------------------------------------------------
    #                                 Pipeline
    #--------------------------------------------------------------------------
    # - .source =   VTKDataSource
    # - .data   =     \_Unstructured_grid
    # - .module =        \_Module_manager
    # - .glyph  =           \__Glyph
    #--------------------------------------------------------------------------

        self._make_data()

        self._make_glyph()

        self._module = plotutilities.make_vector_module(self._color,
                                                      self._opacity,
                                                      self._lut_range,
                                                      self._lut_expansion,
                                                      self._reverse_lut)

        self._assemble_pipeline()

        self.on_scene = False

        @property
        def type(self):
            return 'Vectors-source'
    #--------------------------------------------------------------------------
    #                               Data Structure
    #--------------------------------------------------------------------------

    def _make_data(self):
        try:
            self._vectors.shape
        except:
            self._vectors = np.array(self._vectors)
        if len(self._vectors.shape) == 1:
            self._vectors = np.array([self._vectors])
        norm = np.linalg.norm(self._vectors, axis=1)
        if hasattr(self._anchor, 'vertices'):
            if self._anchor.type == 'Mesh':
                self._geometry = self._anchor
                if self._anchor_mode == 'vertex':
                    points = self._geometry.vertices
                elif self._anchor_mode == 'face':
                    points = self._geometry.face_barycenters()
                elif self._anchor_mode == 'edge':
                    points = self._geometry.edge_mid_points()
            if self._anchor.type == 'Points':
                self._geometry = self._anchor
                points = self._geometry.vertices
        else:
            try:
                self._anchor.shape
            except:
                self._anchor = np.array(self._anchor)
            if len(self._anchor.shape) == 1:
                self._anchor = np.array([self._anchor])
            points = self._anchor
        mask = np.where(norm > 0)[0]
        points = points[mask]
        vectors = self._vectors[mask]
        self._data = tvtk.UnstructuredGrid(points=points)
        if self._scalars is not None:
            scalars = self._scalars[mask]
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = 'scalars'
        self._data.point_data.vectors = vectors
        self._data.point_data.vectors.name = self.name

    #--------------------------------------------------------------------------
    #                                Glyph
    #--------------------------------------------------------------------------

    def _make_glyph(self):

        self._glyph = Glyph()
        self._glyph.glyph.scale_mode = 'scale_by_vector'
        self._glyph.glyph.color_mode = 'color_by_vector'
        self._glyph.actor.property.opacity = self._opacity
        #print(vector.glyph.glyph_source.glyph_dict)

        if self._glyph_type == '3D-arrow':
            g_src = self._glyph.glyph.glyph_source.glyph_dict['arrow_source']
            g_src.shaft_resolution = self._resolution
            g_src.tip_resolution = self._resolution
            if self._shaft_radius != None:
                g_src.shaft_radius = self._shaft_radius
            if self._tip_length != None:
                g_src.tip_length = self._tip_length
            if self._tip_radius != None:
                g_src.tip_radius = self._tip_radius
            self._glyph.glyph.glyph_source.glyph_source = g_src

        elif self._glyph_type == 'cylinder':
            g_src = self._glyph.glyph.glyph_source.glyph_dict['cylinder_source']
            g_src.resolution = self._resolution
            g_src.radius = self._radius
            g_src.height = self._height
            self._glyph.glyph.glyph_source.glyph_source = g_src

        elif self._glyph_type == 'line':
            g_src = self._glyph.glyph.glyph_source.glyph_dict['glyph_source2d']
            g_src.dash = True
            g_src.glyph_type = 'none'
            g_src.scale = self._scale_factor
            #g_src.line_width = self._line_width
            self._glyph.glyph.glyph_source.glyph_source = g_src

        elif self._glyph_type == 'cone':
            self._glyph.actor.actor.property.representation = 'surface'
            g_src = self._glyph.glyph.glyph_source.glyph_dict['cone_source']
            g_src.resolution = self._resolution
            g_src.radius = self._radius
            g_src.height =self._height
            self._glyph.glyph.glyph_source.glyph_source = g_src

        elif self._glyph_type == 'axes':
            g_src = self._glyph.glyph.glyph_source.glyph_dict['axes']
            g_src.line_width = self._line_width
            #g_src.scale_factor = self._scale_factor
            self._glyph.glyph.glyph_source.glyph_source = g_src

        else:
            self._glyph = vtk_vectors()

        self._glyph.glyph.glyph.scale_factor = self._scale_factor
        #self._glyph.trait_modified = True
        self._glyph.glyph.glyph_source.glyph_position = self._position
        if not self._scaling:
            self._glyph.glyph.glyph.scaling = False
        self._glyph.actor.mapper.interpolate_scalars_before_mapping = True
        if self._glossy:
            self._glyph.actor.actor.property.specular = 0.7 * self._glossy
            self._glyph.actor.actor.property.specular_power = 11 * self._glossy
        if not self._shading:
            self._glyph.actor.actor.property.lighting = False

    #--------------------------------------------------------------------------
    #                                Pipeline
    #--------------------------------------------------------------------------

    def _assemble_pipeline(self):
        src = VTKDataSource(data=self._data)
        self._module.add_child(self._glyph)
        src.add_child(self._module)
        self.source = src

    # -------------------------------------------------------------------------
    #                                 Update
    # -------------------------------------------------------------------------

    def _update_lut_range(self, **kwargs):
        lut_range = np.array(self.data_range)
        self._module.vector_lut_manager.data_range = lut_range
        self._module.vector_lut_manager.use_default_range = False

    def _update_glyph(self, **kwargs):
        self._scale_factor = kwargs.get('scale_factor', self._scale_factor)
        self._glyph.glyph.glyph.scale_factor = self._scale_factor

    def _update_data(self, **kwargs):
        self._vectors = kwargs.get('vectors', self._vectors)
        self._scalars = kwargs.get('scalars', self._scalars)
        norm = np.linalg.norm(self._vectors, axis=1)
        mask = np.where(norm > 0)[0]
        self.data_range = [0, np.max(norm)]
        if self._geometry is not None:
            if self._geometry.type == 'Mesh':
                #self._geometry = self._anchor
                if self._anchor_mode == 'vertex':
                    anchor = self._geometry.vertices
                elif self._anchor_mode == 'face':
                    anchor = self._geometry.face_barycenters()
                elif self._anchor_mode == 'edge':
                    anchor = self._geometry.edge_mid_points()
                anchor = anchor[mask]
                self._data.set(points=anchor)
            elif self._anchor.type == 'Points':
                self._geometry = self._anchor
                anchor = self._geometry.vertices
                anchor = anchor[mask]
                self._data.set(points=anchor)
        elif 'anchor' in kwargs:
            anchor = kwargs['anchor']
            anchor = anchor[mask]
            self._data.set(points=anchor)
        vectors = self._vectors[mask]
        if self._scalars is not None:
            self._scalars = np.array(self._scalars)
            scalars = self._scalars[mask]
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = 'scalars'
        self._data.point_data.vectors = vectors
        self._data.point_data.vectors.name = self.name
        self._data.modified()

    # -------------------------------------------------------------------------
    # -------------------------------------------------------------------------

    def update(self, **kwargs):
        self._update_data(**kwargs)
        self._update_lut_range(**kwargs)
        self._update_glyph(**kwargs)
        self.source.update()

    # -------------------------------------------------------------------------
    # -------------------------------------------------------------------------


#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from traits.api import HasTraits, Instance, Property, Enum, Button, String,\
                       on_trait_change, Float, Bool, Int,Constant, ReadOnly,\
                       List, Array, Range, File, observe

from traitsui.api import View, Item, HSplit, VSplit, InstanceEditor, HGroup,\
                         Group, ListEditor, Tabbed, VGroup, CheckListEditor,\
                         ArrayEditor, Action, ToolBar, Separator, Controller,\
                         FileEditor, RangeEditor

from pyface.image_resource import ImageResource

# -----------------------------------------------------------------------------

'''check.py: an interactive checker for mesh connectivity and normals'''

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                Tools Handler
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class T_Handler(Controller):

    def close(self, info, is_ok):
        info.object._closed = True
        Controller.close(self ,info, is_ok)
        return True


class Tool(HasTraits):

    def __init__(self):
        HasTraits.__init__(self)
        self._scenemanager = None

    @property
    def current_object(self):
        return self._scenemanager.current_object

    @property
    def geometry(self):
        return self._scenemanager.current_object.geometry

    @property
    def mesh(self):
        return self._scenemanager.current_object.geometry

    @property
    def meshmanager(self):
        if self._scenemanager.current_object.type != 'Mesh_plot_manager':
            return None
        return self._scenemanager.current_object

    @property
    def scenemanager(self):
        return self._scenemanager

    @property
    def geolab(self):
        return self._scenemanager

    @scenemanager.setter
    def scenemanager(self, scene_manager):
        self._scenemanager = scene_manager

    def close(self):
        try:
            self._closed = True
            self._handler.info.ui.dispose()
        except:
            pass


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                    B-spline
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class BsplineRebuilder(Tool):

    _closed = Bool(True)

    _handler = T_Handler()

    u_sampling = Int(0, label='sampling (U,V)')

    v_sampling = Int(0)

    u_degree = Int(0, label='degree (U,V)')

    v_degree = Int(0)

    u_points = Int(0, label='points (U,V)')

    v_points = Int(0)

    rebuild_button = Button(label='Rebuild')

    resample_button = Button(label='Resample')

    error_string = String('0', label='Deviation')

    view = View(VGroup(HGroup(Item('u_sampling', width=-100),
                                    Item('v_sampling', show_label=False,
                                         width=-100),),
                       HGroup(Item('u_points', width=-100),
                                    Item('v_points', show_label=False,
                                         width=-100),),
                       HGroup(Item('u_degree', width=-100),
                                    Item('v_degree', show_label=False,
                                         width=-100),),
                       HGroup(Item('rebuild_button', show_label=False),
                              Item('resample_button', show_label=False),
                              #'_',
                              #Item('error_string', style='readonly'),
                              ),
                       show_border=True),
                title='B-spline rebuilder',
                handler=_handler,
                width=350,
                icon=ImageResource('../icons/bspline_rebuild.png'))

    @property
    def bspline(self):
        return self.scenemanager.current_object.geometry


    def start(self):
        if not self._closed:
            self._handler.info.ui.dispose()
        if self.bspline.type != 'Bspline_Surface':
            return
        self.u_sampling = self.bspline.u_sampling
        self.v_sampling = self.bspline.v_sampling
        self.u_degree = self.bspline.u_degree
        self.v_degree = self.bspline.v_degree
        P = self.bspline.number_of_control_points
        self.u_points = P[0]
        self.v_points = P[1]
        self._closed = False
        self.configure_traits()

    @on_trait_change('rebuild_button')
    def rebuild_fired(self):
        self.bspline.rebuild(self.u_degree, self.v_degree, self.u_points,
                             self.v_points)
        self.scenemanager.object_changed()
        self.scenemanager.current_object.update_plot()

    @on_trait_change('resample_button')
    def resample_fired(self):
        self.bspline.u_sampling = self.u_sampling
        self.bspline.v_sampling = self.v_sampling
        self.scenemanager.object_changed()
        self.scenemanager.current_object.update_plot()

    @on_trait_change('update_button')
    def update_fired(self):
        self.scenemanager.update_plot()
        self.scenemanager.current_object.update_plot()


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                      Loads
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class Loads(Tool):

    _closed = Bool(True)

    _handler = T_Handler()

    area_load = Float(0, label='Area load')

    beam_load = Float(1, label='Beam load')

    vector = Array(np.float, (1, 3), editor=ArrayEditor(width=20))

    apply_force = Button(label='Apply force')

    reset_forces = Button(label='Reset forces')

    view = View(VGroup('area_load',
                       'beam_load',
                       'vector',
                       HGroup(
                              Item('apply_force',
                                   show_label=False,
                                   resizable=True),
                              Item('reset_forces',
                                   show_label=False,
                                   resizable=True),
                              ),
                       show_border=True),
                title='Loads',
                handler=_handler,
                width=300,
                icon=ImageResource('../icons/applyforce.png'))

    def start(self):
        if not self._closed:
            self._handler.info.ui.dispose()
        self.area_load = self.mesh.area_load
        self.beam_load = self.mesh.beam_load
        self.configure_traits()
        self._closed = False
        self.meshmanager.hide('forces')
        self.plot_loads()

    @on_trait_change('_closed')
    def hide_loads(self):
        self.meshmanager.hide('apply_loads')
        self.meshmanager.update_plot()

    @on_trait_change('area_load')
    def update_area_load(self):
        self.mesh.area_load = self.area_load
        self.plot_loads()

    @on_trait_change('beam_load')
    def update_beam_load(self):
        self.mesh.beam_load = self.beam_load
        self.plot_loads()

    @on_trait_change('apply_force')
    def apply_force_fired(self):
        selected = self.meshmanager.selected_vertices
        self.mesh.apply_force(self.vector[0,:], selected)
        self.plot_loads()

    @on_trait_change('reset_forces')
    def reset_forces_fired(self):
        self.mesh.reset_forces()
        self.plot_loads()

    def plot_loads(self):
        self.meshmanager.plot_vectors(vectors = self.mesh.loads(),
                                      color = 'blue-red',
                                      glyph_type = '3D-arrow',
                                      position = 'head',
                                      name = 'apply_loads')


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                              Incremental Remesh
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

    def start(self):
        if not self._closed:
            self._handler.info.ui.dispose()
        self.configure_traits()
        self._closed = False

    @on_trait_change('corner_tolerance')
    def set_corner_tolerance(self):
        self.mesh.corner_tolerance = self.corner_tolerance

    @on_trait_change('remesh_button')
    def incremental_remesh(self):
        if self.meshmanager is not None:
            self.mesh.incremental_remesh(self.remesh_factor)
            self.geolab.fix_edges_bug()
            self.meshmanager.update_plot()
            self.geolab.object_changed()


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                              Corner Tolerance
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class CornerTolerance(Tool):

    corner_tolerance = Range(0.0, 1.0, 0.3)

    _closed = Bool(True)

    _handler = T_Handler()

    view = View(Group(Item('corner_tolerance',
                            editor = RangeEditor(mode='slider'),),
                      show_border=True),
                title='Set corner tolerance',
                handler=_handler,
                icon=ImageResource('../icons/corners.png'))


    @property
    def scenemanager(self):
        return self._scenemanager

    @scenemanager.setter
    def scenemanager(self, scene_manager):
        self._scenemanager = scene_manager
        ct = (scene_manager.current_object.mesh.corner_tolerance + 1)/2
        self.corner_tolerance = ct

    def start(self):
        if not self._closed:
            self._handler.info.ui.dispose()
        self.configure_traits()
        self._closed = False
        self.plot_corners()
        self.scenemanager.current_object.add_plot_callback(self.plot_corners,
                                                           name='corners')

    def close(self):
        try:
            self._closed = True
            self._handler.info.ui.dispose()
        except:
            pass

    @on_trait_change('corner_tolerance')
    def set_corner_tolerance(self):
        self.mesh.corner_tolerance = self.corner_tolerance*2 - 1
        self.plot_corners()

    @on_trait_change('_closed')
    def hide_corners(self):
        self.scenemanager.current_object.hide('corners')
        self.scenemanager.current_object.remove_plot_callback('corners')

    def plot_corners(self):
        corners = self.mesh.mesh_corners()
        r = self.scenemanager.current_object.r
        self.scenemanager.current_object.plot_glyph(vertex_indices=corners,
                                                  glyph_type = 'sphere',
                                                  radius=2.2*r,
                                                  color='g',
                                                  name='corners')


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                   Save Mesh
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

class S_Handler(Controller):

    def close(self, info, is_ok):
        info.object._closed = True
        if info.initialized:
            info.object.current_object.highlight_off()
        Controller.close(self ,info, is_ok)
        return True


class SaveGeometry(Tool):

    _closed = Bool(True)

    _handler = S_Handler()

    label = String('opt')

    save_button = Button(label='Save')

    view = View(HGroup('label',
                       Item('save_button',
                            show_label=False),
                       show_border=True),
                title='Save geometry',
                handler=_handler,
                icon=ImageResource('../icons/save.png'))

    def start(self):
        if not self._closed:
            self._handler.info.ui.dispose()
        self.configure_traits()
        self._closed = False
        self.current_object.highlight()

    def close(self):     
        try:
            self._closed = True
            self._handler.info.ui.dispose()
        except:
            pass

    @observe('save_button')
    def save_file(self, event):
        name = ('{}_{}').format(self.geometry.name, self.label)
        self.scenemanager.object_save(None)
        path = self.geometry.make_obj_file(name)
        self.current_object.highlight_off()
        self.scenemanager.object_saved(path)
        self.close()







# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

#------------------------------------------------------------------------------

__author__ = 'Davide Pellis'

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------

def bspline_basis_functions(t, degree, knot_vector):
    m = len(knot_vector) - degree - 2
    N = np.zeros((len(t), m + 2))
    for i in range(m):
        c1 = t >= knot_vector[i]
        c2 = t < knot_vector[i+1]
        ind = np.logical_and(c1, c2)
        N[ind,i] = 1
    c1 = t >= knot_vector[m]
    c2 = t <= knot_vector[m+1]
    ind = np.logical_and(c1, c2)
    N[ind,m] = 1
    for r in range(1, degree + 1):
        for j in range(m + 1):
            a = knot_vector[j+r] - knot_vector[j]
            if a != 0:
                A = (t - knot_vector[j]) / a
            else:
                A = np.zeros(len(t))
            b = knot_vector[j+r+1] - knot_vector[j+1]
            if b != 0:
                B = (knot_vector[j+r+1] - t) / b
            else:
                B = np.zeros(len(t))
            N[:,j] = A*N[:,j] + B*N[:,j+1]
    N = np.delete(N, -1, axis=1)
    return N

def bspline_derivative_basis_functions(t, degree, knot_vector, d=1):
    m = len(knot_vector) - degree - 2
    if degree < d:
        return np.zeros((len(t), m + 1))
    N = bspline_basis_functions(t, degree-d, knot_vector[d:-d])
    for n in range(d):
        i = np.arange(m+1 - (d-n-1)) + (d-n-1)
        a = (knot_vector[i + degree - (d-n-1)] - knot_vector[i])
        zero = np.where(a == 0)[0]
        a[zero] = 1
        f = (degree - (d-n-1)) / a
        f[zero] = 0
        f = np.hstack((f,[0]))
        N = np.column_stack((np.zeros(len(N)), N, np.zeros(len(N))))
        N = np.einsum('j,ij->ij', f, N)
        N = N[:,0:-1] - N[:,1:]
    return N




#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from tvtk.api import tvtk

from mayavi.sources.vtk_data_source import VTKDataSource

from mayavi.modules.surface import Surface

from mayavi.filters.tube import Tube

import numpy as np

#------------------------------------------------------------------------------

from wsurf.geolab.plotter import plotutilities

# -----------------------------------------------------------------------------

'''polylinesource.py: The plot source for Polylines and Circles'''

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                   Polyline
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

class Polyline(object):

    def __init__(self, polyline, **kwargs):

        self._polyline      = polyline

        self._edge_data      = kwargs.get('edge_data',      None)

        self._vertex_data    = kwargs.get('vertex_data',    None)

        self._color          = kwargs.get('color',          'cornflower')

        self._opacity        = kwargs.get('opacity',        1)

        self._line_width     = kwargs.get('line_width',     1)

        self._tube_radius    = kwargs.get('tube_radius',    None)

        self._tube_sides     = kwargs.get('tube_sides',     10)

        self._lut_range      = kwargs.get('lut_range',      '-:0:+')

        self._lut_expansion  = kwargs.get('lut_expansion',  2)

        self._reverse_lut    = kwargs.get('reverse_lut',    False)

        self._shading        = kwargs.get('shading',        True)

        self._glossy         = kwargs.get('glossy',         0.3)

        self._sampling       = kwargs.get('sampling',       50)

        self.name            = kwargs.get('name',           'edges')

        self._data_range     = None



    #--------------------------------------------------------------------------
    #                                 Pipeline
    #--------------------------------------------------------------------------
    # - .source  =   VTKDataSource
    # - .data    =     \_Poly_data
    # - .tube    =       None  \_Tube
    # - .module  =        \_______\_Module_manager
    # - .surface =                   \__Surface
    #--------------------------------------------------------------------------

        self._make_data()

        self._make_tube()

        self._module = plotutilities.make_module(self._color,
                                                self._opacity,
                                                self._lut_range,
                                                self._lut_expansion,
                                                self._reverse_lut,
                                                self._data_range)
        self._make_surface()

        self._assemble_pipeline()

        self.on_scene = False

    @property
    def type(self):
        return 'Polyline-source'

    #--------------------------------------------------------------------------
    #                             Data Structure
    #--------------------------------------------------------------------------

    def _make_data(self):
        if self._polyline.type is 'Circle':
            self._polyline.sampling = self._sampling
            self._polyline.make_vertices()
        points = self._polyline.vertices
        cells = self._polyline.cell_array()
        N = cells.shape[0]
        cell_array = tvtk.CellArray()
        cell_array.set_cells(N,cells)
        self._data = tvtk.PolyData(points=points, lines=cell_array)
        if self._edge_data is not None:
            scalars = np.array(self._edge_data)
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [np.min(scalars),np.max(scalars)]
        elif self._vertex_data is not None:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars),np.max(scalars)]
        else:
            scalars = np.zeros([self._polyline.E])
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [0,0]

    #--------------------------------------------------------------------------
    #                                Surface
    #--------------------------------------------------------------------------

    def _make_surface(self):
        self._surface = Surface()
        if self._line_width != None:
            self._surface.actor.property.line_width = self._line_width
        if type(self._opacity) == int or type(self._opacity) == float:
            self._surface.actor.property.opacity = self._opacity
        if not self._shading:
            self._surface.actor.actor.property.lighting = False
        if self._glossy!= 0:
            self._surface.actor.actor.property.specular = 0.7 * self._glossy
            self._surface.actor.actor.property.specular_power = 11*self._glossy

    #--------------------------------------------------------------------------
    #                                 Tube
    #--------------------------------------------------------------------------

    def _make_tube(self):
        self._tube = None
        if self._tube_radius == 'adaptive':
            self._surface.actor.actor.property.representation = 'wireframe'
            self._surface.actor.actor.property.render_lines_as_tubes = True
        elif self._tube_radius is not None:
            self._tube = Tube()
            self._tube.filter.radius = self._tube_radius
            self._tube.filter.number_of_sides = self._tube_sides
            self._tube

    #--------------------------------------------------------------------------
    #                               Pipeline
    #--------------------------------------------------------------------------

    def _assemble_pipeline(self):
        src = VTKDataSource(data=self._data)
        self._module.add_child(self._surface)
        if self._tube is None:
            src.add_child(self._module)
        else:
            self._tube.add_child(self._module)
            src.add_child(self._tube)
        self.source = src

    # -------------------------------------------------------------------------
    #                                 Update
    # -------------------------------------------------------------------------

    def _update_data(self, **kwargs):
        if True:
            self._polyline = kwargs.get('polyline', self._polyline)
            if self._polyline.type is 'Circle':
                self._polyline.sampling = self._sampling
                self._polyline.make_vertices()
            points = self._polyline.vertices
            cells = self._polyline.cell_array()
            N = cells.shape[0]
            cell_array = tvtk.CellArray()
            cell_array.set_cells(N, cells)
            cell_array = tvtk.CellArray()
            cell_array.set_cells(self._polyline.E, cells)
            self._data.set(lines=cell_array)
            self._data.set(points=points)
        self._edge_data = kwargs.get('edge_data', None)
        self._vertex_data = kwargs.get('vertex_data', None)
        if self._edge_data is not None:
            scalars = np.array(self._edge_data)
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [np.min(scalars),np.max(scalars)]
        elif self._vertex_data is not None:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars),np.max(scalars)]
        else:
            scalars = np.zeros([self._polyline.E])
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [0,0]
        self._data.modified()

    def _update_lut_range(self, **kwargs):
        self._lut_range = kwargs.get('lut_range', self._lut_range)
        lut_range = plotutilities.make_lut_range(self._lut_range, self._data_range)
        self._module.scalar_lut_manager.data_range = lut_range

    def _update_color(self, **kwargs):
        if 'color' in kwargs:
            if kwargs['color'] != self._color:
                self._color = kwargs['color']
                plotutilities.make_lut_table(self._module,
                                             self._color,
                                             self._opacity,
                                             self._lut_expansion,
                                             self._reverse_lut)

    #--------------------------------------------------------------------------
    #--------------------------------------------------------------------------

    def update(self, **kwargs):
        self._update_data(**kwargs)
        self._update_lut_range(**kwargs)
        self._update_color(**kwargs)
        self.source.update()

    #--------------------------------------------------------------------------
    #--------------------------------------------------------------------------



# -*- coding: utf-8 -*-


#------------------------------------------------------------------------------
#                                   HANDLER
#------------------------------------------------------------------------------


class Handler(object):

    def __init__(self):

        self.handled = {}

        self.state_callbacks = []

        self.state_variables = {}


    def add_state_callback(self, callback):
        self.state_callbacks.append(callback)

    def set_state(self, state=None):
        for name in self.handled:
            try:
                self.handled[name]._set_state(state)
            except:
                pass
        for callback in self.state_callbacks:
            try:
                callback(state)
            except:
                pass

    def get_variable(self, name, default=None):
        try:
            variable = self.state_variables[name]
            return variable()
        except:
            variable = self.state_variables[name]
            return variable()

    def add_variable(self, name, value):
        self.state_variables[name] = value




#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from scipy import sparse

from scipy import spatial

from scipy.sparse import linalg

# -----------------------------------------------------------------------------

from wsurf.geolab import utilities

# -----------------------------------------------------------------------------

'''polyine.py: The Polyline data structure class'''

__author__ = 'Davide Pellis'


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
#                                   POLYLINE
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

class Polyline(object):

    def __init__(self, vertices, closed=False):

        self.name = 'polyline'

        if type(vertices) == str:
            self.read_obj_file(vertices)
        else:
            self.vertices = vertices

        self.closed = closed

        self.corner_tolerance = None

        self._kdtree = None

    # --------------------------------------------------------------------------
    #                               Attributes
    # --------------------------------------------------------------------------

    @property
    def type(self):
        return 'Polyline'

    @property
    def V(self):
        return self.vertices.shape[0]

    @property
    def E(self):
        if self.closed:
            return self.V
        else:
            return self.V - 1

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        vertices = np.array(vertices, 'f')
        if len(vertices.shape) != 2 or vertices.shape[1] != 3:
            raise ValueError('wrong size!')
        try:
            self._vertices[:, :] = vertices[:, :]
        except:
            self._vertices = vertices
        self._kdtree = None

    @property
    def corner_tolerance(self):
        return self._corner_tolerance

    @corner_tolerance.setter
    def corner_tolerance(self, corner_tolerance):
        if corner_tolerance is None:
            self._corner_tolerance = None
        else:
            self._corner_tolerance = float(corner_tolerance)

    @property
    def closed(self):
        return self._closed

    @closed.setter
    def closed(self, bool):
        self._closed = bool

    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------

    def __str__(self):
        if self.closed:
            out = 'Closed '
        else:
            out = 'Open '
        out += ('polyline: |V| = {}').format(self.V)

        return out

    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------

    def read_obj_file(self, file_name):
        file_name = str(file_name)
        self.name = file_name.split('.')[0]
        obj_file = open(file_name, encoding='utf-8')
        vertices_list = []
        for l in obj_file:
            splited_line = l.split(' ')
            if splited_line[0] == 'v':
                split_x = splited_line[1].split('\n')
                x = float(split_x[0])
                split_y = splited_line[2].split('\n')
                y = float(split_y[0])
                split_z = splited_line[3].split('\n')
                try:
                    z = float(split_z[0])
                except ValueError:
                    print('WARNING: disable line wrap when saving .obj')
                vertices_list.append([x, y, z])
        self.vertices = np.array(vertices_list)

    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------

    def vertex_tangents(self, normalized=False):
        i = np.arange(self.V)
        v1 = np.roll(i, 1)
        v2 = np.roll(i, -1)
        if not self.closed:
            v1[0] = 0
            v2[-1] = self.V - 1
        if normalized:
            N = self.vertices[v2, :] - self.vertices[v1, :]
        else:
            E1 = self.vertices[v2, :] - self.vertices[i, :]
            E2 = self.vertices[i, :] - self.vertices[v1, :]
            N = utilities.normalize(E1) + utilities.normalize(E2)
        N = utilities.normalize(N)
        return N

    def vertex_normals(self):
        from geolab.optimization.combnormals import comb_normals
        N = comb_normals(self)
        return N

    def vertex_bisectors(self, return_cosines=False):
        v0 = np.arange(self.V)
        v2 = np.roll(v0, 1)
        v1 = np.roll(v0, -1)
        V1 = self.vertices[v1, :] - self.vertices[v0, :]
        V2 = self.vertices[v0, :] - self.vertices[v2, :]
        V1 = utilities.normalize(V1)
        V2 = utilities.normalize(V2)
        B = V1 - V2
        B = utilities.normalize(B)
        if not self.closed:
            N0 = np.cross(np.cross(V1[0], B[1, :]), V1[0])
            B[0, :] = N0 / (np.linalg.norm(N0) + 1e-10)
            N1 = np.cross(np.cross(V2[-1], B[-2, :]), V2[-1])
            B[-1, :] = N1 / (np.linalg.norm(N1) + 1e-10)
        if not return_cosines:
            return B
        else:
            A = np.einsum('ij,ij->i', V1, -V2)
            if not self.closed:
                A[0] = A[-1] = -1
            return B, A

    def _pipe_vectors(self):
        return self.vertex_bisectors(return_cosines=False)

    def corners(self):
        if self.corner_tolerance is None:
            return []
        v1 = np.arange(self.V)
        v0 = np.roll(v1, 1)
        v2 = np.roll(v1, -1)
        if not self.closed:
            v0[0] = self.V - 1
            v2[-1] = 0
        V1 = self.vertices[v2, :] - self.vertices[v1, :]
        V1 = utilities.normalize(V1)
        V2 = self.vertices[v1, :] - self.vertices[v0, :]
        V2 = utilities.normalize(V2)
        C = np.einsum('ij,ij->i', V1, V2)
        corners = v1[np.where(C[:] < self.corner_tolerance)[0]].tolist()
        return corners

    def refine(self, steps=5):
        if steps <= 0:
            return
        V = self.V
        corners = self.corners()
        if self.closed:
            N = steps * V
        else:
            N = steps * (V - 1)
        v = np.arange(V)
        s = np.arange(V + N)
        d = np.repeat(3, V + N)
        d1 = np.repeat(-2, V + N)
        d2 = np.repeat(0.5, V + N)
        d[(steps + 1) * v] = 1
        d1[(steps + 1) * v] = 0
        d2[(steps + 1) * v] = 0
        i = np.hstack((s, s, s, s, s))
        j = np.hstack((s, np.roll(s, 1), np.roll(s, -1),
                       np.roll(s, 2), np.roll(s, -2)))
        data = np.hstack((d, d1, d1, d2, d2))
        if not self.closed:
            data[1] = data[V + N - 2] = 2.5
            data[V + N + 1] = data[3 * (V + N) - 2] = -1
            data[3 * (V + N) + 1] = 0
            data[5 * (V + N) - 2] = 0
            if corners is not None and len(corners) > 0:
                corners = np.array(corners)
                corners = np.delete(corners, np.where(corners == 0)[0])
        if corners is not None and len(corners) > 0:
            c = (steps + 1) * np.array(corners)
            data[np.roll(s, 1)[c]] = 2.5
            data[np.roll(s, -1)[c]] = 2.5
            data[V + N + np.roll(s, -1)[c]] = -1
            data[2 * (V + N) + np.roll(s, 1)[c]] = -1
            data[3 * (V + N) + np.roll(s, -1)[c]] = 0
            data[4 * (V + N) + np.roll(s, 1)[c]] = 0
        M = sparse.coo_matrix((data, (i, j)), shape=(V + N, V + N))
        P = np.zeros((N + V, 3))
        P[(steps + 1) * v] = self.vertices
        M = sparse.csc_matrix(M)
        X = linalg.spsolve(M, P[:, 0], use_umfpack=False)
        Y = linalg.spsolve(M, P[:, 1], use_umfpack=False)
        Z = linalg.spsolve(M, P[:, 2], use_umfpack=False)
        P = np.vstack((X, Y, Z)).T
        self.vertices = P

    # ------------------------------------------------------------------------------
    #                                  CLOSENESS
    # ------------------------------------------------------------------------------

    def make_kdtree(self):
        KDtree = spatial.cKDTree(self.vertices)
        self._kdtree = KDtree

    def closest_vertices(self, points, make_tree=False):
        if self._kdtree is None:
            self.make_kdtree()
        elif make_tree:
            self.make_kdtree()
        closest = self._kdtree.query(points)[1]
        return closest

    # ------------------------------------------------------------------------------
    #                                VISUALIZATION
    # ------------------------------------------------------------------------------

    def cell_array(self, offset=0):
        vi = np.arange(self.V) + offset
        vj = np.roll(vi, -1) + offset
        c = np.repeat(2, self.vertices.shape[0])
        cells = np.vstack((c, vi, vj)).T
        if not self.closed:
            cells = np.delete(cells, self.V - 1, axis=0)
        cells = np.ravel(cells)
        return cells


if __name__ == '__main__':
    from geolab import plotter

    P = Polyline(10 * np.random.random((10, 3)))
    P.corner_tolerance = 1
    P.refine(2)

    plot = []

    plot.append(plotter.Polyline(P, color='r'))
    V = P._pipe_vectors()
    plot.append(plotter.Vectors(V, anchor=P.vertices))
    plotter.view(plot)



# -*- coding: utf-8 -*-

from wsurf.geolab.utilities.utilities import sum_repeated, repeated_range,\
                                       orthogonal_vectors, orthogonal_vector,\
                                       normalize, remap

from wsurf.geolab.utilities.stringutilities import print_progress, make_filepath,\
                                             serial

from wsurf.geolab.utilities.bsplineutilities import bspline_basis_functions,\
                                     bspline_derivative_basis_functions

